/*
** mime.c - written by vesely in milano on 26 sep 1999
**
** current version only parses binary input from HTTP transport
**
** TODO:
** - bit stuffing unencode
** - set content-id
** - rfc2049 headers encoding utility
** - - content-description
** - cut and paste (mime_clip_tag)
*/
/*
** Part of MIMEC: see COPYING
** Copyright (c) 1999, 2000, 2001, 2002, 2003 Alessandro Vesely
** All rights reserved.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>

#if !defined(NDEBUG) && defined(INCLUDE_DMALLOC)
#define DMALLOC_FUNC_CHECK
#include <dmalloc.h>
#endif

#if defined(TEST_MAIN)
#include <errno.h>
#endif

#if !defined(SIZE_MAX)
#define SIZE_MAX UINT_MAX
#endif

#include "mime.h"
#include "parse.h"
#include "charencode.h" /* generated by do_charencode, contains char_encode */

/* -----------------------------------------------------
** STRUCTURE DEFINITIONS
*/
typedef struct mmap_area
{
	void* start;
	size_t length;
	enum {
		mmap_area_empty,
		mmap_area_malloced,
		mmap_area_mapped} flag;
	content_ptr adjusted;
} mmap_area;

typedef struct mime_entity_tag
{
	unsigned char *boundary;
	unsigned char *description;
	unsigned char *content_type;
	unsigned char *name;
	unsigned char *filename;
	
	mmap_area mmarea; /* -ale: 24sep2003: set if content is memory mapped */
		
	FILE *fpw;	/* client's file */
	long offset, length, actual_max_line_length;
	long hoffset, hlength; /* -ale: 24sep2003: set if header is saved */
	int internal_id;
	
	unsigned charset:4; /* 0=not-set, 1=us-ascii, n>1=iso-8859-(n-1) */
	unsigned dispo:2;
	unsigned name_from_dispo:1;
	unsigned is_last_part:1;
	unsigned is_leaf_part:1;
	unsigned is_discrete:1;
	unsigned is_alt_file:1;
	unsigned is_for_discard:1;
	unsigned error_base64:1;
	unsigned error_quoted_printable:1;
	unsigned error_system:1;
	unsigned error_structure:1; /* -ale: 19dec2001 */
	unsigned warning_structure:1; /* -ale: 16feb2004 */
	unsigned mime_type:4;
	unsigned mime_subtype:4;
	unsigned encoding:4;
	unsigned actual_encoding:4;
	
	unsigned found_term_boundary:1; /* -ale: 11may2002 */
} MIME_ENTITY;

struct mime_tag
{
	FILE *fp, *fpw;
	unsigned char *buffer;
	size_t buf_max; /* currently this is constant */
	unsigned char *buf_first, *buf_end;
	
	MIME_ENTITY *ment;
	size_t ment_max, ment_cur, ment_count, ment_committed;
	
	unsigned default_encoding;
	unsigned default_charset;
	unsigned sys_pagesize; // system pagesize
	long max_data_size;
	char crlf[4];
	
	char boundary_rand[8];
	char boundary_unique[20];
	unsigned long no_boundary[2]; /* 64 bits */
	unsigned boundary_ndx;
	
	int next_internal_id;
	mime_error last_error;
	unsigned header_u_count; /* learn unix_line_break */
	unsigned header_b_count; /* learn binary line break */
	unsigned unix_line_break:1; /* set on reads */
	unsigned dot_stuffing:1; /* done only on writes */
	unsigned seek_stuck:1;
	unsigned save_headers:1;
};

struct mime_clip_tag
{
	MIME *m;
	MIME_ENTITY *ment;
	size_t ment_count;
};

/* WARNING: the following must match mime_encoding_enum
** they are in order of decreasing variety
** mime_7bit is the first that doesn't need conversion on receive (ident)
*/
static const char *mime_encoding_desc[] =
{
	"()", /* not set */
	"base64",
	"quoted-printable",
	"7bit",
	"8bit",
	"binary"
};

#if !defined(BITSPERBYTE)
#define	BITSPERBYTE	8
#endif
#if !defined(BITS)
#define	BITS(type)	(BITSPERBYTE * (long)sizeof (type))
#endif


/* boundary generation: use 64 chars to generate boundary_unique */
static const char allowed_boundary_64[] = /* missing "'()+-_.,?<>" */
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789=:";
typedef int my_compile_time_check[
	(BITS(((MIME*)0)->no_boundary) == sizeof allowed_boundary_64 -1) ? 1 : -1];

/* format arguments: 1) boundary_rand, 2) any number, 3) boundary_unique
** boundary_rand (7) + boundary_unique (20) + template (17) = 44
** still allows for 70 - 44 = 26 chars of part number
**                                      123  456790123  4567  */
static const char *boundary_template = "---%s---(part_%d)---%s";

static unsigned char bin2base64_tab[]=
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static unsigned char const base64_tab[] =
{
	 -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   /*   0 -   7 */
	 -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   /*   8 -  15 */
	 -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   /*  16 -  23 */
	 -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   /*  24 -  31 */
	 -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   /*  32 -  39 */
	 -1,  -1,  -1,  62,  -1,  -1,  -1,  63,   /*  40 -  47 */
	 52,  53,  54,  55,  56,  57,  58,  59,   /*  48 -  55 */
	 60,  61,  -1,  -1,  -1,  -1,  -1,  -1,   /*  56 -  63 */
	 -1,   0,   1,   2,   3,   4,   5,   6,   /*  64 -  71 */
	  7,   8,   9,  10,  11,  12,  13,  14,   /*  72 -  79 */
	 15,  16,  17,  18,  19,  20,  21,  22,   /*  80 -  87 */
	 23,  24,  25,  -1,  -1,  -1,  -1,  -1,   /*  88 -  95 */
	 -1,  26,  27,  28,  29,  30,  31,  32,   /*  96 - 103 */
	 33,  34,  35,  36,  37,  38,  39,  40,   /* 104 - 111 */
	 41,  42,  43,  44,  45,  46,  47,  48,   /* 112 - 119 */
	 49,  50,  51,  -1,  -1,  -1,  -1,  -1    /* 120 - 127 */

};

#define MAX_LINE_LENGTH 76 /* excluding trailing CRLF */
#define MAX_BOUNDARY_LENGTH (MAX_LINE_LENGTH + MAX_LINE_LENGTH)
#define MAX_BUFFERED_AHEAD (MAX_BOUNDARY_LENGTH + 6)

#define MIME_SET_HEADER_OK 0
#define MIME_SET_HEADER_ERROR -1
#define MIME_SET_HEADER_IGNORE 2
typedef int (*mime_header_func)(MIME_ENTITY*, PARSE*);
/*
** mime header func:
** if second arg is NULL just reset relevant entity fields
** otherwise set relevant field values and return MIME_HEADER_OK.
** return -1 on (hard) error or 2 ignoring non-relevant garbage
*/
static int mime_set_version(MIME_ENTITY*, PARSE*);
static int mime_set_type(MIME_ENTITY*, PARSE*);
static int mime_set_disposition_internal(MIME_ENTITY*, PARSE*);
static int mime_set_encoding_internal(MIME_ENTITY*, PARSE*);
static int mime_set_description_internal(MIME_ENTITY*, PARSE*);
static int mime_set_id(MIME_ENTITY*, PARSE*);


typedef void (*mime_write_func)(MIME_ENTITY*, char const *h, FILE*, char const *);
/*
** mime_write_header: called with header name,
** writes its field and terminates with CRLF
*/
static void mime_write_version(MIME_ENTITY*, char const *h, FILE*, char const *);
static void mime_write_type(MIME_ENTITY*, char const *h, FILE*, char const *);
static void mime_write_disposition(MIME_ENTITY*, char const *h, FILE*, char const *);
static void mime_write_encoding(MIME_ENTITY*, char const *h, FILE*, char const *);

struct mime_field_name_tag
{
	char const *name;
	mime_header_func fn;
	mime_write_func fnw;
};

/* WARNING: in the following,
** the version must be the first header (see mime_write_header_internal)
** and the order of fields must be that of mime_header_enum
*/
static const struct mime_field_name_tag mime_field_name[] =
{
	{"MIME-Version", mime_set_version, mime_write_version},
	{"Content-Type", mime_set_type, mime_write_type},
	{"Content-Disposition", mime_set_disposition_internal, mime_write_disposition},
	{"Content-Transfer-Encoding", mime_set_encoding_internal, mime_write_encoding},
	{"Content-Description", mime_set_description_internal, NULL},
	{"Content-ID", mime_set_id, NULL}
};


/*
* silent for automated tests
*/
#if defined(TEST_MAIN)
static int be_silent = 0;
#endif


/* -----------------------------------------------------
** MIME_ENTITY
*/


static int mime_entity_unmap(MIME_ENTITY* const ment)
{
	int rtc = 0;
	mmap_area *const mmarea = &ment->mmarea;
	switch (mmarea->flag)
	{
		default:
		case mmap_area_empty:
			break;
		case mmap_area_malloced:
			free(mmarea->start);
			break;
		case mmap_area_mapped:
			rtc = munmap(mmarea->start, mmarea->length);
			break;
	}
	mmarea->start = NULL;
	mmarea->length = 0;
	mmarea->flag = mmap_area_empty;
	return rtc;
}

static void mime_entity_clear(MIME_ENTITY* const ment)
{
	assert(ment);
	
	if (ment->boundary)
		free(ment->boundary);
	if (ment->description)
		free(ment->description);
	if (ment->content_type)
		free(ment->content_type);
	if (ment->name)
		free(ment->name), ment->name = NULL;
	if (ment->filename)
		free(ment->filename), ment->filename = NULL;
	if (ment->mmarea.flag != mmap_area_empty)
		mime_entity_unmap(ment);
	/*
	if (ment->fpw)
	** do nothing: fp_alt is opened and closed by caller
	*/
	memset(ment, 0, sizeof *ment);
}

static void mime_entity_init(MIME_ENTITY* const ment, int internal_id)
{
	memset(ment, 0, sizeof *ment);
	ment->internal_id = internal_id;
	ment->is_discrete = 1;
}

static void mime_init_multipart_mixed(MIME_ENTITY* const ment, int internal_id)
{
	memset(ment, 0, sizeof *ment);
	ment->mime_type = mime_multipart;
	ment->internal_id = internal_id;
	/* already memset: ment->is_discrete = 0; */
	ment->mime_subtype = mime_multipart_mixed;
}

#if defined(REPLACE_FPUTS)
static size_t length_fputs(char const *s, FILE *fp)
{
	if (fputs(s, fp))
		return strlen(s);
	else
		return 0;
}
#else
static size_t length_fputs(char const *s, FILE *fp)
{
	return fputs(s, fp);
}
#endif

static size_t print_header_param(FILE* fp, size_t length, char const *crlf,
	char const *fmt, ...)
/*
** write header parameter, preceded by " ; ". insert line breaks when needed
** assume no line breaks already coded
*/
{
	char buf[128];
	va_list stack;
	size_t nlength;

	va_start(stack, fmt);
	nlength = vsnprintf(buf, sizeof buf, fmt, stack);
	
	if (nlength + 1 >= sizeof buf)
	{             /*1   12   */
		fprintf(fp, " ;%s  ", crlf);
		length = 2 + vfprintf(fp, fmt, stack);
	}
	else
	{
		if (length + 3 + nlength > MAX_LINE_LENGTH)
		{             /*1   12   */
			fprintf(fp, " ;%s  ", crlf);
			length = 2;
		}
		else
			length += length_fputs(" ; ", fp);
			
		length += length_fputs(buf, fp);
	}
	
	va_end(stack);
	return length;
}

static int mime_set_version(MIME_ENTITY* const ment, PARSE *p)
{
	/* trust is 1.0 */
	(void)ment;
	(void)p;
	return MIME_SET_HEADER_OK;
}

static void mime_write_version(MIME_ENTITY* const ment, char const *h,
	FILE *fp, char const * const crlf)
{
	(void)ment;
	fprintf(fp, "%s: 1.0%s", h, crlf);
}

static int mime_set_type(MIME_ENTITY* const ment, PARSE *p)
/* this may change entity role, setting is_discrete according to field */
{
	int rtc = MIME_SET_HEADER_ERROR, sep;
	char *mtype, *msubtype;

	/*
	** if found once more or p==NULL, clear previous values
	*/
	if (ment->boundary)
		free(ment->boundary), ment->boundary = NULL;
	if (ment->content_type)
		free(ment->content_type), ment->content_type = NULL;
	if (ment->name && ment->name_from_dispo == 0)
		free(ment->name), ment->name = NULL;
	ment->charset = 0;

	if (p &&
		get_mime_word(p, &mtype, &sep) && mtype != NULL)
	{
		if (sep == 0)
		/* -ale:26sep2001 Netscape6 writes"unknown" on empty files */
		{
			ment->mime_type = mime_application;
			ment->mime_subtype = mime_application_octetstream;
			if ((ment->content_type = strdup(mtype)) != 0)
				rtc = MIME_SET_HEADER_OK;
		}
		else if (sep == '/' &&
			get_mime_word(p, &msubtype, &sep) != 0 &&
			msubtype != NULL &&
			(ment->content_type = (unsigned char*)
				malloc(strlen(mtype) + strlen(msubtype) + 2)) != NULL)
		{
			enum mime_type_enum mime_type;
			char *pname, *pval;
		
			strcpy(ment->content_type, mtype);
			strcat(ment->content_type, "/");
			strcat(ment->content_type, msubtype);

			if (stricmp(mtype, "multipart") == 0)
			/*
			** recognize multipart/form-data or multipart/mixed
			*/
			{
				ment->mime_type = mime_type = mime_multipart;
				ment->is_discrete = 0;
				if (stricmp(msubtype, "form-data") == 0)
					ment->mime_subtype = mime_multipart_form_data;
				else
					ment->mime_subtype = mime_multipart_mixed;
			}
			else if (stricmp(mtype, "text") == 0)
			/*
			** for text, assume text/plain and get charset parameter
			** (fall back to application/octetstream if not recognized
			** charset, but leave textual content as originally coded)
			**
			** if charset is not specified, leave undefined (transp. default)
			*/
			{
				ment->is_discrete = 1;
				ment->mime_type = mime_type = mime_text;
				if (stricmp(msubtype, "plain") == 0)
					ment->mime_subtype = mime_text_plain;
				else
					ment->mime_subtype = mime_text_other;
			}
			else if (stricmp(mtype, "message") == 0)
			{
				ment->is_discrete = 1;
				ment->mime_type = mime_type = mime_message;
				if (stricmp(msubtype, "rfc822") == 0)
				{
					ment->mime_subtype = mime_message_rfc822;
					ment->is_discrete = 0;
				}
				else if (stricmp(msubtype, "delivery-status") == 0 ||
					stricmp(msubtype, "disposition-notification") == 0 ||
					stricmp(msubtype, "rfc822-headers") == 0)
						ment->mime_subtype = mime_message_machine;
				else
					ment->mime_subtype = mime_message_other;
			}
			else
			{
				ment->is_discrete = 1;
				ment->mime_type = mime_type = mime_application;
				ment->mime_subtype = mime_application_octetstream;
			}
			
			while (sep == ';' && 
				get_mime_word_relaxed(p, &pname, &sep) &&
				sep == '=' && pname != NULL)
			{
				if (get_mime_word_relaxed(p, &pval, &sep) == 0 || pval == NULL)
				/* -ale: 22oct2003: allow "content-type: text/plain; charset="
				** as set by qmail */
				{
					sep = 0;
					break;
				}
				
				if (stricmp(pname, "name") == 0)
				/*
				** accept name=blah unless name set on
				** content-disposition in form-data
				*/
				{
					if (ment->name_from_dispo == 0 &&
						ment->name == NULL &&
						(ment->name = strdup(pval)) == 0)
							ment->error_system = 1;
				}
				else switch (mime_type)
				{
					case mime_multipart:
						if (stricmp(pname, "boundary") == 0)
						{
							if (ment->boundary == NULL &&
								(ment->boundary = strdup(pval)) == NULL)
									ment->error_system = 1;
						}
						break;
					
					case mime_text:
						if (stricmp(pname, "charset") == 0)
						{
							if (stricmp(pval, "us-ascii") == 0)
								ment->charset = 1;
							else if (strincmp(pval, "iso-8859-", 9) == 0)
							{
								char *t;
								unsigned long l = strtoul(pval + 9, &t, 10);
								if (t == NULL || *t != 0 || l > 15)
								{
									ment->mime_type = mime_application;
									ment->mime_subtype = mime_application_octetstream;
								}
								else
									ment->charset = l + 1;
							}
						}
						break;
						
					default:
						break;
				}
			}
			/* -ale: 24nov99 not doing boundary check,
			** called from set_content_type */
			
			/* -ale: 25oct01 accept terminating ';' (amazon) */
			if (sep == 0 || sep == ';')
				rtc = MIME_SET_HEADER_OK;
		}
	}
		
	return rtc;
}

static void set_content_type_string(MIME_ENTITY* const ment)
/*
**
*/
{
	if (ment->content_type == NULL)
		switch (ment->mime_type)
		{
			case mime_text:
				ment->content_type = strdup("text/plain");
				break;
				
			case mime_multipart:
				if (ment->mime_subtype == mime_multipart_form_data)
					ment->content_type = strdup("multipart/form-data");
				else
					ment->content_type = strdup("multipart/mixed");
				break;
				
			case mime_application:
			default:
				ment->content_type = strdup("application/octet-stream");
				break;
		}
	if (ment->content_type == NULL)
		ment->error_system = 1;
}

static void mime_write_type(MIME_ENTITY* const ment, char const *h,
	FILE *fp, char const * const crlf)
{
	set_content_type_string(ment);
	if (ment->content_type != NULL)
	{
		size_t length = fprintf(fp, "%s: %s", h, ment->content_type);
		switch (ment->mime_type)
		{
			case mime_text:
				if (ment->charset == 1)
					length = print_header_param(fp, length, crlf,
						"%s", "charset=US-ASCII");
				else if (ment->charset > 1)
					length = print_header_param(fp, length, crlf,
						"charset=iso-8859-%d", ment->charset - 1);
				break;
				
			case mime_multipart:
				if (ment->boundary)
					length = print_header_param(fp, length, crlf,
						"boundary=\"%s\"", ment->boundary);
				break;
				
			case mime_application:
			default:
				break;
		}
		if (ment->name != NULL && ment->dispo != mime_dispo_is_form_data_field)
			print_header_param(fp, length, crlf, "name=\"%s\"", ment->name);
		fputs(crlf, fp);
	}
}

static int mime_set_disposition_internal(MIME_ENTITY* const ment, PARSE *p)
{
	/* -ale: 12nov2001 invalid disposition is ignored */
	int rtc = MIME_SET_HEADER_ERROR, sep;
	char *mdispo;
	
	/*
	** if found once more or en==NULL, clear previous values
	*/
	if (ment->filename)
		free(ment->filename), ment->filename = NULL;
	if (ment->name && ment->name_from_dispo)
		free(ment->name), ment->name = NULL, ment->name_from_dispo = 0;
	ment->dispo = mime_dispo_not_set;
	
	if (p)
	{
		rtc = MIME_SET_HEADER_IGNORE, sep;

		/*
		** for form-data disposition, must get name parameter
		*/
		if (get_mime_word_relaxed(p, &mdispo, &sep) && mdispo != NULL)
		{
			char *pname, *pval;

			if (stricmp(mdispo, "form-data") == 0)
				ment->dispo = mime_dispo_is_form_data_field;
			else if (stricmp(mdispo, "inline") == 0)
				ment->dispo = mime_dispo_is_inline;
			else if (stricmp(mdispo, "attachment") == 0)
				ment->dispo = mime_dispo_is_attachment;
			else /* -ale:17sep2001 for Sircam attachment:
			     ** "Content-Disposition: message text" */
				sep = 'A';
			
			while (sep == ';' && 
				get_mime_word_relaxed(p, &pname, &sep) &&
				sep == '=' && pname != NULL &&
				get_mime_word_relaxed(p, &pval, &sep) && pval != NULL)
			{
				if (stricmp(pname, "name") == 0)
				{
					if (ment->name && ment->name_from_dispo == 0)
						free(ment->name), ment->name = NULL;
					if (ment->name == NULL)
					{
						if ((ment->name = strdup(pval)) != NULL)
							ment->name_from_dispo = 1;
						else
							ment->error_system = 1;
					}
				}
				else if (stricmp(pname, "filename") == 0)
				{
					if (ment->filename == NULL &&
						(ment->filename = strdup(pval)) == NULL)
							ment->error_system = 1;
				}
			}
			
			/* -ale: 12nov01 accept terminating ';'  */
			if ((sep == 0 || sep == ';') &&
				(ment->dispo != mime_dispo_is_form_data_field || /* name needed for cgi */
					ment->name != NULL))
						rtc = MIME_SET_HEADER_OK;
		}
	}
	
	return rtc;
}

static void mime_write_disposition(MIME_ENTITY *const ment, char const *h,
	FILE *fp, char const * const crlf)
{
	size_t length;
	
	if (ment->dispo == mime_dispo_not_set &&
		(ment->name == NULL || ment->dispo != mime_dispo_is_form_data_field) &&
		ment->filename == NULL)
			return;
			
	length = fprintf(fp, "%s: ", h);
	switch (ment->dispo)
	{
		case mime_dispo_is_form_data_field:
			length += length_fputs("form-data", fp);
			break;
		case mime_dispo_is_inline:
			length += length_fputs("inline", fp);
			break;
		case mime_dispo_is_attachment:
			length += length_fputs("attachment", fp);
			break;
		default:
			if (ment->filename)
				length += length_fputs("attachment", fp);
			break;
	}
		
	if (ment->name && ment->dispo == mime_dispo_is_form_data_field)
		length = print_header_param(fp, length, crlf,
			"name=\"%s\"", ment->name);
	if (ment->filename)
		length = print_header_param(fp, length, crlf,
			"filename=\"%s\"", ment->filename);
	fputs(crlf, fp);
}

static int mime_set_encoding_internal(MIME_ENTITY* const ment, PARSE *p)
{
	int rtc = MIME_SET_HEADER_ERROR, sep;
	char *mencode;
	
	/*
	** if found once more or en==NULL, clear previous values
	*/
	ment->encoding = 0;
	
	if (p)
	{
		/* -ale: bugfix: 28jan02: set default encoding to mime_7bit */
		rtc = MIME_SET_HEADER_IGNORE;
		ment->encoding = mime_7bit;
		if (get_mime_word_relaxed(p, &mencode, &sep) && mencode != NULL)
		{
			size_t i;
			
			for (i = 0;
				i < sizeof mime_encoding_desc/sizeof mime_encoding_desc[0]; ++i)
					if (stricmp(mencode, mime_encoding_desc[i]) == 0)
					{
						ment->encoding = i;
						rtc = MIME_SET_HEADER_OK;
						break;
					}
		}
	}
	
	return rtc;
}

static void mime_write_encoding(MIME_ENTITY* const ment, char const *h,
	FILE *fp, char const * const crlf)
{
	fprintf(fp, "%s: %s%s", h, mime_encoding_desc[ment->encoding], crlf);
}

static int mime_set_description_internal(MIME_ENTITY* const ment, PARSE *p)
{
	char const *en = "TODO: parse description";
	/*
	** if found once more or en==NULL, clear previous values
	*/
	if (ment->description)
		free(ment->description), ment->description = NULL;

	(void)p;

	/* TODO: check for rfc2047-encoded field body */
	if (en && ment->description == NULL &&
		(ment->description = strdup(en)) == NULL)
			ment->error_system = 1;
	
	return MIME_SET_HEADER_OK;
}

static int mime_set_id(MIME_ENTITY* const ment, PARSE *p)
{
	/* TODO */
	(void)ment;
	(void)p;
	return MIME_SET_HEADER_OK;
}

static int internal_mime_set_header_internal(PARSE *p, MIME_ENTITY *const ment,
	mime_header_func fn, char const *hint)
{
	int rtc = 0, sep;
	char *name;
	size_t i_fn = sizeof mime_field_name / sizeof mime_field_name[0];
	
	parse_set_no_hash_comment(p, 1);
	
	if (fn == NULL)
	{
		if (get_mime_word(p, &name, &sep) && name != NULL && sep == ':')
		{
			for (i_fn = 0; i_fn < sizeof mime_field_name / sizeof mime_field_name[0]; ++i_fn)
				if (stricmp(name, mime_field_name[i_fn].name) == 0)
				{
					fn = mime_field_name[i_fn].fn;
					break;
				}
		}
		//else -ale: cannot inhibit headers like '>Return-path: blah'
		//	rtc = -1;
	}

	if (fn != NULL)
	{
		if ((rtc = (*fn)(ment, p)) != MIME_SET_HEADER_OK)
		{
			(*fn)(ment, NULL); /* reset any relevant field */
			if (rtc != MIME_SET_HEADER_IGNORE)
				rtc = -1;
		}
		else
			rtc = 1;
	}
	/* else unknown header */

#if defined(TEST_MAIN)
	if (rtc < 0)
	{
		fprintf(stderr, "rtc=%d from %s", rtc, hint);
		if (i_fn < sizeof mime_field_name / sizeof mime_field_name[0])
			fprintf(stderr, " fn=%s ", mime_field_name[i_fn].name);
		fputc('\n', stderr);
	}
#else
	(void)hint;
#endif

	return rtc;
}

static int mime_set_header_internal(MIME_ENTITY *const ment, char const *en,
	mime_header_func fn)
/*
** return 0 if header ok but not recognized, 1 if ok, -1 if error
** invalid header (no column) yield an error -16oct2003.
** this may change entity role, setting is_discrete according to the
** Content-Type field.
*/
{
	PARSE *p;
	
	if (en && (p = parse_init_memory(en, NULL, -1L, NULL)) != NULL)
	{
		int rtc = internal_mime_set_header_internal(p, ment, fn, en);
		parse_clear(p);
		return rtc;
	}
	
	ment->error_system = 1;
	return -1; /* hard err */
}

static int mime_set_header_internal_from_file(MIME_ENTITY *const ment, FILE *tmpf,
	mime_header_func fn)
/*
** return 0 if header ok but not recognized, 1 if ok, -1 if error
** invalid header (no column) yield an error -16oct2003.
** this may change entity role, setting is_discrete according to the
** Content-Type field.
*/
{
	if (tmpf)
	{
		PARSE *p;
		long size = ftell(tmpf);
		if (size >= 0 && fseek(tmpf, 0, SEEK_SET) == 0 &&
			(p = parse_init_file(tmpf, "tmpfile", size, NULL)) != NULL)
		{
			int rtc = internal_mime_set_header_internal(p, ment, fn, "tmpfile");
			parse_clear(p);
			return rtc;
		}
	}
	
	ment->error_system = 1;
	return -1; /* hard err */
}

static void mime_write_header_internal(MIME_ENTITY* const ment, FILE *fp,
	char const * const crlf, unsigned from)
/*
** if from == 0 write also the version, otherwise start from 1
*/
{
	if (from)
		from = 1;
		
	while (from < sizeof mime_field_name / sizeof mime_field_name[0])
	{
		mime_write_func const fnw = mime_field_name[from].fnw;
		if (fnw)
			(*fnw)(ment, mime_field_name[from].name, fp, crlf);
		++from;
	}
}


/* -----------------------------------------------------
** MIME GENERAL
*/

#if !defined(NDEBUG)
/* check function, (see ttree.c) */
static unsigned reach_last_node(MIME const * const m, unsigned i)
{
	if (i >= m->ment_count)
		return -1;
	
	for (; i < m->ment_count; ++i)
	{
		MIME_ENTITY const * const ment = m->ment + i;
		if (!ment->is_leaf_part)
		{
			i = reach_last_node(m, i + 1);
			if ((int)i == -1)
				return -1;
		}
		if (ment->is_last_part)
			return i;
	}
	
	assert(0);
	return -1;
}

static int mime_invariant(MIME const * const m)
{
	/*
	** check structure consistency
	*/
	int rtc = m != NULL &&
		m->ment != NULL && m->ment_max >= m->ment_count &&
			m->ment_cur < m->ment_count &&
		(m->ment_committed == (size_t) -1 ||
			m->ment_committed < m->ment_count) &&
		m->buffer != NULL &&
		m->buf_max > 0 &&
		m->buf_first >= m->buffer &&
		m->buf_first <= m->buffer + m->buf_max &&
		m->buf_end >= m->buffer &&
		m->buf_end <= m->buffer + m->buf_max &&
		m->buf_first <= m->buf_end &&
		m->fpw != NULL &&
		m->next_internal_id > 0;
	
	/*
	** check buffer minimal capacity (this should be compile-time
	** until buf_max is constant).
	** Test required for statements such as
	**    if (buf_first + SOMETHING > buf_end &&
	**       buf_end - buffer >= buf_max)
	**          fill_buffer()
	** where SOMETHING is less than MAX_BUFFERED_AHEAD
	*/
	if (rtc && m->buf_max <= MAX_BUFFERED_AHEAD)
		rtc = 0;
	
	if (rtc && m->ment_count > 0) /* tree consistency check */
	{
		int balance;
		unsigned i;

		/*
		** the root node must have no siblings
		*/
		if (!m->ment[0].is_last_part)
			return 0;
		
		/*
		** among one node's children, exactly one must be the last
		** (these guarantees return from next recursive check)
		*/
		balance = 1;
		for (i = 0; i < m->ment_count; ++i)
		{
			MIME_ENTITY const * const ment = m->ment + i;
			unsigned j;
			int internal_id = ment->internal_id;
			
			if (ment->is_last_part)
				if (--balance < 0)
					return 0;
			if (!ment->is_leaf_part)
				++balance;
			
			if (internal_id <= 0 || internal_id >= m->next_internal_id)
				return 0;
			
			for (j = 0; j < i; ++j)
				if (m->ment[j].internal_id == internal_id)
					return 0;
		}
		if (balance != 0)
			return 0;

		/*
		** recursively check traversability
		*/
		return reach_last_node(m, 0) + 1 == m->ment_count;
	}
	
	return rtc;
}
#else
static const int mime_invariant(MIME const * const m) { return 1; }
#endif /* NDEBUG */

static size_t fill_buffer(MIME* const m)
{
	size_t size;

	assert(mime_invariant(m));
	
	if (m->buf_end > m->buf_first) /* some data not read yet */
	{
		size_t const size_in = m->buf_end - m->buf_first;
		if (m->buf_first > m->buffer)
		{
			memmove(m->buffer, m->buf_first, size_in);
			m->buf_first = m->buffer;
			m->buf_end = m->buf_first + size_in;
		}
		size = m->buf_max - size_in;
	}
	else
	{
		size = m->buf_max;
		m->buf_first = m->buf_end = m->buffer;
	}

	if (feof(m->fp) || ferror(m->fp))
		size = 0;
		
	if (size > 0)
	{
		size = fread(m->buf_end, 1, size, m->fp);
		m->buf_end += size;
	}
	
	if (ferror(m->fp))
		m->last_error = MIME_INFILE_ERROR;
	
	assert(mime_invariant(m));
	assert(m->buf_end >= m->buf_first);

	return m->buf_end - m->buf_first;
}

static int discard_from_buffer(MIME* const m, unsigned char *p)
/*
* discard the character pointed by p from the buffer.
* return non-zero if the character was the last in the buffer,
* return 0 if p now points to the next valid char.
*/
{
	int rtc;

	assert(mime_invariant(m));
	assert(p >= m->buf_first);
	assert(p < m->buf_end);
	
	if (p < --m->buf_end)
	{
		size_t const size = m->buf_end - p;
		memmove(p, p + 1, size);
		rtc = 0;
	}
	else rtc = 1;

	assert(mime_invariant(m));	
	return rtc;
}

static int check_buffer(MIME* const m, size_t min_size)
/* return non zero if min_size available */
{
	size_t const size_in = m->buf_end - m->buf_first;
	
	assert(mime_invariant(m));
	
	if (size_in < min_size)
		return fill_buffer(m) >= min_size;

	return 1;
}

static int boundary_may_fit_in_buffer(MIME *m)
// check buffer contains at least "--x" but possibly enlarge
{
	size_t const size_in = m->buf_end - m->buf_first;
	
	assert(mime_invariant(m));
	
	if (size_in > MAX_LINE_LENGTH || feof(m->fp))
		return size_in >= 3;
	return fill_buffer(m) >= 3;
}

void mime_clear(MIME *m)
{
	size_t i;
	
	if (m->fpw)
		fclose(m->fpw);
	if (m->buffer)
		free(m->buffer);

	if (m->ment)
	{
		for (i = 0; i < m->ment_count; ++i)
			mime_entity_clear(m->ment + i);
		free(m->ment);
	}
	
	free(m);
}

MIME* mime_new(enum mime_encoding_enum default_encoding,
	enum mime_charset_enum default_charset)
/*
** prepare new mime struct with default encoding and charset
*/
{
	unsigned boundary_rand, i;
	MIME* const m = (MIME*)malloc(sizeof(MIME));
	
	if (m)
	{
		memset(m, 0, sizeof *m);
		if ((m->ment = (MIME_ENTITY*)
			malloc((m->ment_max = 4) * sizeof(MIME_ENTITY))) == NULL)
		{
			free(m);
			return NULL;
		}

#if defined(TEST_MAIN)
		m->buf_max = 2048;
#else		
		if ((m->buf_max = getpagesize()) < 4096)
			m->buf_max = 4096;
#endif
		
		if ((m->buffer = (unsigned char*)malloc(m->buf_max)) == NULL)
		{
			mime_clear(m);
			return NULL;
		}
		
		if ((m->fpw = tmpfile()) == NULL)
		{
			mime_clear(m);
			return NULL;
		}		
		m->default_encoding = default_encoding;
		m->default_charset = default_charset;

		/* before calling fill_buffer, set buf_end = buffer */
		m->buf_first = m->buffer;
		m->buf_end = m->buffer + m->buf_max;
	
		strcpy(m->crlf, "\r\n");
		m->ment_committed = -1;
		m->ment_count = 1;
		m->next_internal_id = 2;
		mime_entity_init(m->ment, 1);
		m->ment->is_last_part = 1;
		m->ment->is_leaf_part = 1;
		
		srand(time(0));
		boundary_rand = rand() * (ULONG_MAX/RAND_MAX) + rand();
		for (i = 0; i < sizeof m->boundary_rand - 1; ++i)
		{
			unsigned ch = boundary_rand & 0x3f; /* 6 bits */
			m->boundary_rand[i] = allowed_boundary_64[ch];
			boundary_rand >>= 6;
		}
		/* already memset: m->boundary_rand[i] = 0 */
		
		m->sys_pagesize = getpagesize();
		
		assert(mime_invariant(m));
		assert(m->ment_cur == 0);
		assert(m->ment_count == 1); /* top level entity defined */
	}
	return m;
}

enum mime_charset_enum mime_get_default_charset(MIME const * const m)
	{ assert(mime_invariant(m));
	return m->default_charset; }
enum mime_charset_enum mime_set_default_charset(MIME* const m,
	enum mime_charset_enum ch)
	{ assert(mime_invariant(m));
	return m->default_charset = ch; }
mime_error mime_get_last_error(MIME const*const m)
	{ assert(mime_invariant(m));
	return m->last_error; }
	
char const *mime_get_last_error_string(MIME const*const m)
{
	assert(mime_invariant(m));
	switch (m->last_error)
	{
		case MIME_NO_ERROR:         return "no error";
		case MIME_CALL_ERROR:       return "program error";
		case MIME_BUFFER_ERROR:     return "buffer too small";
		case MIME_DATA_ERROR:       return "bad mime input data";
		case MIME_HEADER_ERROR:     return "cannot parse mime header";
		case MIME_INTERNAL_ERROR:   return "internal error";
		case MIME_INFILE_ERROR:     return "input data fault";
		case MIME_FILE_ERROR:       return "internal file corrupted";
		case MIME_SYSFILE_ERROR:    return "system file error";
		case MIME_OVERFLOW_ERROR:   return "too many entities";
		case MIME_SYSTEM_ERROR:     return "system failure";
		default:                    return "corrupted internal structure";
	}
}

static int 
recurse_level(MIME const * const m, size_t i, size_t const target,
	int level, int *result)
{
	assert(result);
	assert((int)i != -1);
	assert(i <= target);
	
	for (; i < target; ++i)
	{
		MIME_ENTITY const * const ment = m->ment + i;
		if (!ment->is_leaf_part)
		{
			i = recurse_level(m, i + 1, target, level + 1, result);
			if ((int)i == -1)
				return -1;
		}
		if (ment->is_last_part)
			return i;
	}
	
	assert(i == target);
	assert(level >= 0);
	
	*result = level;
	return -1;
}

static int get_level(MIME const *const m, size_t target)
{
	int result = -1;
	int i;
	
	assert(mime_invariant(m));
	assert(target < m->ment_count);
	
	i = recurse_level(m, 0, target, 0, &result);

	assert(mime_invariant(m));
	assert(i == -1);
	assert(result >= 0);
	return result;
}

static int left_or_up(MIME const * const m, size_t i,
	size_t target, int *result, mime_seek_arg const a)
/*
** walk down from top of tree to target,
** set result to index of parent or last sibling
** (possibly same as target)
*/
{
	size_t left = target;
	size_t const up = i > 0 ? i - 1 : 0;
	
	assert(a == mime_seek_left || a == mime_seek_up);
	/* assert(mime_invariant(m)); */
	assert(target < m->ment_count);
	assert(i <= target);
	assert(result != NULL);

	for (; i < target; ++i)
	{
		MIME_ENTITY const * const ment = m->ment + i;
		
		left = i;
		if (!ment->is_leaf_part)
			i = left_or_up(m, i + 1, target, result, a);

		if (ment->is_last_part)
			return i;
	}
	
	if (i == target)
		*result = a == mime_seek_up ? up : left;
		
	return m->ment_count;
}

static int seek_last_child(MIME const * const m, size_t i);
static int seek_last_sibling(MIME const * const m, size_t i)
/*
** return index of last sibling
*/
{
	MIME_ENTITY const* ment;
	assert(mime_invariant(m));
	
	while (!(ment = m->ment + i)->is_last_part)
	{
		if (!ment->is_leaf_part)
			i = seek_last_child(m, i);
		++i;
	}

	assert(i < m->ment_count);
	return i;
}

static int seek_last_child(MIME const * const m, size_t i)
/*
** return index of last element in subtree
*/
{
	MIME_ENTITY const* ment;
	assert(mime_invariant(m));
	
	if (!(ment = m->ment + i)->is_leaf_part)
		i = seek_last_child(m,
			seek_last_sibling(m, i + 1));

	assert(i < m->ment_count);
	return i;
}



/* -----------------------------------------------------
** MIME READ
*/

static int discard_upto_crlf(MIME* const m)
/*
** return 0 on successfully discarded CRLF, -1 on error
*/
{
	assert(mime_invariant(m));

	while (check_buffer(m, 1))
	{
		unsigned char *p = m->buf_first, *const e = m->buf_end;
		
		while (p < e)
		{
			int const ch = *p++;
			if (ch == '\r' || ch == '\n')
			{
				m->buf_first = p;
				
				if (ch == '\r')
					if (check_buffer(m, 1) &&
						*m->buf_first == '\n')
							m->buf_first += 1;
					else
						break;
						
				assert(mime_invariant(m));
				return 0;
			}
		}

		if (p == e)
			m->buf_first = p;
	}
	
	if (m->last_error != MIME_INFILE_ERROR)
		m->last_error = MIME_DATA_ERROR;
	assert(mime_invariant(m));
	return -1;
}

static char const* find_end_of_line(MIME const* const m)
/*
** assume buffer already filled (e.g. by read_mime_header)
** and return end of line or NULL
*/
{
	char const *p = m->buf_first;
	char const * const e = m->unix_line_break ? m->buf_end : m->buf_end - 1;
	int const eol = m->unix_line_break ? '\n' : '\r';
	
	for (; p < e; ++p)
		if (*p == eol &&
			(eol == '\n' || p[1] == '\n'))
				return p;
	
	return NULL;
}

#if 0
static char * find_start_of_line(MIME* const m)
/*
** assume buffer already filled (e.g. by read_mime_header)
** and return start of line or NULL
*/
{
	unsigned char *p = m->buf_first;
	unsigned char * const e = m->buffer;
	
	if (p < m->buf_end)
		for (; p >= e; --p)
			if (*p == '\n')
			{
				if (m->unix_line_break)
					return p;
				if (--p >= e && *p == '\r')
					return p;
				++p;
			}
	
	return NULL;
}
#endif

static int set_boundary_unique(MIME* const m, unsigned ndx, int ch)
/*
** ndx is the index into boundary_unique, ch the candidate's char there
** return non-zero if not interested in further chars
**
** implement a simple greedy algorithm to get a unique boundary
** enlarging unique_boundary when forced to do so;
*/
{
	int i_it, i_other, val;
	unsigned long mask;
	char const *const into_val = strchr(allowed_boundary_64, ch);
	
	assert(ndx <= m->boundary_ndx);
	assert(sizeof m->no_boundary / sizeof m->no_boundary[0] == 2);
	assert(m->no_boundary[0] != ULONG_MAX || m->no_boundary[1] != ULONG_MAX);
	
	if (into_val == NULL)
		return 1;
		
	if (ndx < m->boundary_ndx)
		return 0;
	
	val = into_val - &allowed_boundary_64[0];
	i_it = val / BITS(m->no_boundary[0]);
	i_other = !i_it;
	mask = 1 << val % BITS(m->no_boundary[0]);
	
	assert(i_it == 0 && i_other == 1 || i_it == 1 && i_other == 0);
		
	/* if still inside an already seen subsequence, not interested */
	if ((m->no_boundary[i_it] & mask) != 0)
		return 1;
	
	/* if must enlarge unique_boundary, get next char */
	if ((m->no_boundary[i_it] |= mask) == ULONG_MAX &&
		m->no_boundary[i_other] == ULONG_MAX)
	{
		if (ndx < sizeof m->boundary_unique - 1) /* leave trailing 0 */
			m->boundary_unique[ndx] = ch;
		++m->boundary_ndx;
		m->no_boundary[0] = m->no_boundary[1] = 0;
		return 0;
	}
	
	/* otherwise don't need to enlarge */
	return 1;
}

static int commit_boundary_unique(MIME* const m)
{
	unsigned const ndx = m->boundary_ndx;
	unsigned i, j;
	
	if (ndx < sizeof m->boundary_unique - 1)
		for (i = 0; i < sizeof m->no_boundary / sizeof m->no_boundary[0]; ++i)
			for (j = 0; j < BITS(m->no_boundary[0]); ++j)
			{
				unsigned long const mask = 1 << j;
				if ((m->no_boundary[i] & mask) == 0)
				{
					int const val = i * BITS(m->no_boundary[0]) + j;
					m->boundary_unique[ndx] = allowed_boundary_64[val];
					return 0;
				}
			}
			
	return -1;
}

typedef enum check_boundary_mode
	{ check_min, check_max} check_boundary_mode;
	
static int check_boundary_candidate(MIME* const m,
	char const *candidate, unsigned *avail, check_boundary_mode mode)
/*
** assume candidate was found in the body after CRLF + "--"
**
** check if a known boundary in [0, cur) matches the
** candidate and return that index in case. avail is
** assigned the actual length of the matched bundary.
**
** otherwise return -1 and adjust our notion of unique boundary
*/
{
	int i_cur;
	size_t i_template, i_rand, i_number, i_unique;
	size_t const cur = m->ment_cur;
	char const *p, *end = candidate + *avail;
	enum { b_template, b_rand, b_number, b_unique} b_match;

	assert(mime_invariant(m));
	assert(candidate < end);

	for (i_cur = cur - 1; i_cur >= 0; --i_cur)
	/* -ale: bugfix: 28jan02: was forward; 16feb2004: added mode */
	{
		MIME_ENTITY const * const ment = m->ment + i_cur;
		if (!ment->is_discrete && ment->boundary)
		{
			size_t const size = strlen(ment->boundary);
			if (candidate + size <= end &&
				memcmp(ment->boundary, candidate, size) == 0 &&
				(mode == check_max || seek_last_child(m, i_cur) >= cur))
			{
				*avail = size;
				return i_cur;
			}
		}
	}
	
	if (mode != check_max)
		return -1;
	

	/*
	** looks like a boundary but it isn't
	** format arguments: 1) boundary_rand, 2) any number, 3) boundary_unique
	*/
	i_template = i_rand = i_number = i_unique = 0;
	b_match = b_template;
	for (p = candidate; p < end;)
	{
		switch (b_match)
		{
			case b_template:
			{
				int ch = boundary_template[i_template++];
				if (ch == '%')
				{
					ch = boundary_template[i_template++];
					if (ch == 'd')
						b_match = b_number;
					else if (i_rand == 0)
						b_match = b_rand;
					else
						b_match = b_unique;
					break;
				}
				if (ch == 0 || ch != *p++)
					return -1;
				break;
			}
			case b_number:
			{
				int const ch = *p;
				if (!isdigit(ch) || ++i_number > 10)
					b_match = b_template;
				else
					++p;
				break;
			}
			case b_rand:
			{
				int const ch = m->boundary_rand[i_rand++];
				if (ch == 0)
					b_match = b_template;
				else if (ch != *p++)
					return -1;
				break;
			}
			case b_unique:
			{
				int const ch = *p++;
				if (set_boundary_unique(m, i_unique++, ch))
					return -1;
				break;
			}
		}
	}

	return -1;
}

static long copy_file_addline(FILE *src, FILE *dst)
/*
** copy content o src to dst and add a newline
** return total size written or -1
*/
{
	char buf[4096];
	long written = 0;
	size_t in;
	while ((in = fread(buf, 1, sizeof buf, src)) > 0)
		if (fwrite(buf, in, 1, dst) == 1)
			written += in;
	if (!ferror(src) && !ferror(dst) &&
		fputc('\n', dst) != EOF)
			return written + 1;
	return EOF;
}


static int read_mime_header(MIME* const m, int save_header)
/* 
** Assume buffer is at the beginning of field name.
** The unix_line_break flag is set if most headers have a plain LF termination.
** Spurious CRs may be dropped from the headers.
** Return > 0 if header found, set warning_structure for bad headers.
** Return 0 if end of headers, i.e. empty line, end of file, boundary.
** Return < 0 on fatal error.
*/
{
	size_t size;
	unsigned char *p, *set_buf_first = NULL;
	size_t crlf_size = 2, header_lines;
	FILE *tmpf = NULL;

	int rtc, bad_chars = 0;

	assert(mime_invariant(m));

	/*
	** check for end of headers
	*/
	size = fill_buffer(m);
	if (size < 1) /* no room for a CRLF */
	{
		/* 19dec2001: was:
		if (m->last_error == 0)
			m->last_error = MIME_HEADER_ERROR;
		return -1;
		*/
		return 0;
	}

	/* check empty line for end of fields */
	switch (*m->buf_first)
	{
		case '\n':
			m->buf_first += 1;
			m->unix_line_break =
				++m->header_u_count > m->header_b_count;
			return 0;
		
		case '\r':
			rtc = size > 1 && m->buf_first[1] == '\n' ? 0 : -1;
			if (rtc == -1)
			{
				++m->buf_first; /* discard spurious CR */
				m->ment[m->ment_cur].warning_structure = 1;
				m->last_error = MIME_HEADER_ERROR;
			}
			else
			{
				m->buf_first += 2;
				m->unix_line_break =
					m->header_u_count > ++m->header_b_count;
			}

			assert(mime_invariant(m));
			return 0;

		case '-': /* -ale:16oct2003: end of body? */
			p = m->buf_first;
			if (size > 2 && p[1] == '-')
			{
				unsigned avail = size - 2, blen;
				MIME_ENTITY* const ment = m->ment + m->ment_cur;
				
				p += 2;
				rtc = check_boundary_candidate(m, p, &avail, check_min);
				if (rtc >= 0)
				{
					ment->found_term_boundary = 1;
					if (avail > size)
						avail = size;
					m->buf_first = p + avail;
					m->ment_cur = rtc;
					return 0;
				}
				
				/* first boundary in the headers */
				if (ment->is_discrete == 0 &&
					ment->boundary &&
					(blen = strlen(ment->boundary)) <= avail &&
					(blen == avail || p[blen] == '\r' || p[blen] == '\n') &&
					memcmp(ment->boundary, p, blen) == 0)
				{
					ment->found_term_boundary = 1;
					return 0;
				}
			}
		
		default:
			break;
	}
	
	/*
	** find end of field. 22oct2003: added external loop with tmp file
	** we must return 1, unless fatal error, as it is not end of headers
	*/
	header_lines = 0;
	rtc = -1;
	for (;;)
	{
		for (p = m->buf_first; p < m->buf_end; ++p)
		{
			int ch = *p;
			if (ch == '\n' || ch == '\r')
			{
				++header_lines;
				crlf_size = (ch == '\n') ? 1 : 2;
				if (p + crlf_size >= m->buf_end)
				{
					if (p <= m->buf_first)
					/* end of file implies end of header */
					{
						*p = 0;
						rtc = 1;
						crlf_size = size;
					}
					break;
				}
				
				if (ch == '\n' || (ch == '\r' && p[1] == '\n'))
				{
					if (p[crlf_size] != ' ' && p[crlf_size] != '\t')
					{
						*p = 0;
						rtc = 1; /* end of header found */
						if (ch == '\n')
							++m->header_u_count;
						else
							++m->header_b_count;
						m->unix_line_break = m->header_u_count > m->header_b_count;
						break;
					}
				}
				else /* discard spurious CR */
					ch = 0;
			}
			
			if (ch == 0) /* discard bad char, either spurious CR or
			                bare 0 unmanageable in parse field body */
			{
#if defined(TEST_MAIN)
				fprintf(stderr,
					"Bad char in header '%.*s'\n",
						p - m->buf_first, m->buf_first);
#endif
				m->last_error = MIME_HEADER_ERROR;
				++bad_chars;
				discard_from_buffer(m, p);
				--p;
				--size;
			}
		}
				
		if (rtc > 0) /* header found */
		{
			if (tmpf && p > m->buf_first &&
				fwrite(m->buf_first, p - m->buf_first, 1, tmpf) != 1)
					rtc = -1;
			break;
		}
		
#if defined(TEST_MAIN)
		fprintf(stderr,
			"refilling buffer at line %d of long header\n", header_lines);
#endif

		assert(rtc < 0);
		assert(m->buf_first <= p && p <= m->buf_end);

		if (p > m->buf_first)
		{
			if (tmpf == NULL)
			{
				tmpf = tmpfile();
				if (tmpf == NULL)
					break;
			}
		
			if (fwrite(m->buf_first, p - m->buf_first, 1, tmpf) != 1)
				break;

			m->buf_first = p;
		}

		if ((size = fill_buffer(m)) <= 0)
		{
			rtc = 1; /* header w/o line term */
			p = m->buf_first;
			crlf_size = 0;
			break;
		}
	}
	
	if (rtc < 0)
	{
#if defined(TEST_MAIN)
		fprintf(stderr, "Problem with tmpfile: %s\n", strerror(errno));
#endif
		rtc = -1;
		m->last_error = MIME_SYSTEM_ERROR;
		if (tmpf)
		{
			fclose(tmpf);
			tmpf = NULL;
		}
	}

	set_buf_first = p; /* in case end of line was not found just consume data */
		
	if (rtc >= 0)
	{
		unsigned char * const q = m->buf_first;
		MIME_ENTITY* const ment = m->ment + m->ment_cur;

		if (tmpf)
			rtc = mime_set_header_internal_from_file(ment, tmpf, NULL);
		else
			rtc = mime_set_header_internal(ment, q, NULL);

		set_buf_first = p + crlf_size; /* discard field body and CRLF */
		
		if (rtc == 0 && save_header) /* save not recognized headers */
		{
			long offset;
			mime_error me = MIME_NO_ERROR;
			FILE *const fp = m->fpw;
			
			if (fp == NULL ||
				ferror(fp) ||
				fseek(fp, 0, SEEK_END) != 0 ||
				(offset = ftell(fp)) < 0)
					me = MIME_SYSFILE_ERROR;
			else if (ment->hlength == 0)
				ment->hoffset = offset;
			else if (ment->hoffset + ment->hlength < offset)
				me = MIME_FILE_ERROR;
		
			if (me == MIME_NO_ERROR)
			{
				if (tmpf)
				{
					long hlength;
					if (fseek(tmpf, 0, SEEK_SET) == 0 &&
						(hlength = copy_file_addline(tmpf, fp)) >= 0)
							ment->hlength += hlength;
					else
					{
						me = MIME_SYSFILE_ERROR;
						ment->error_system = 1;
					}
				}
				else if (fwrite(q, p - q, 1, fp) != 1 ||
					fputc('\n', fp) == EOF)
				{
					me = MIME_SYSFILE_ERROR;
					ment->error_system = 1;
				}
				else
					ment->hlength += p - q + 1;
			}
			
			if (me != MIME_NO_ERROR)
			{
				m->last_error = me;
				ment->error_system = 1;
			}
		}

		if (rtc < 0 || bad_chars > 0)
		{
			m->last_error = MIME_HEADER_ERROR;
			ment->warning_structure = 1;
		}
		rtc = ment->error_system ? -1 : 1;
	}
		
#if defined(TEST_MAIN)
	if (rtc < 0 && !be_silent)
	{
		int string_size;
		
		if (set_buf_first)
			string_size = set_buf_first - m->buf_first - crlf_size;
		else
			string_size = size > 40 ? 40 : size;
		fprintf(stderr, "rtc=%d from read_mime_header at \"%.*s\"\n",
			rtc, string_size, m->buf_first);
	}
#endif	
	if (set_buf_first)
		m->buf_first = set_buf_first;
	if (tmpf)
		fclose(tmpf);
	assert(mime_invariant(m));
	return rtc;
}

#if 0 /* not used */
static size_t max_boundary_size(MIME const * const m)
/*
** get the size of the longest boundary in [0, cur)
*/
{
	size_t const cur = m->ment_cur;
	size_t b_size_max = 0, i;

	assert(mime_invariant(m));

	for (i = 0; i < cur; ++i)
	{
		MIME_ENTITY const * const ment = m->ment + i;
		if (!ment->is_discrete)
		{
			size_t const size = strlen(ment->boundary);
			if (size > b_size_max)
				b_size_max = size;
		}
	}
	
	return b_size_max;
}
#endif /* not used */

static size_t sniff_data(MIME* const m, size_t length,
	unsigned char const * const first, unsigned char const * const end)
/*
** check the features of the data in [first, end)
** and return the residual length of line (to
** be passed back in further calls)
**
** data descriptions in current entity are updated
*/
{
	size_t const cur = m->ment_cur;
	MIME_ENTITY* const ment = m->ment + cur;
	unsigned actual_encoding = ment->actual_encoding;
	long actual_max_line_length = ment->actual_max_line_length;
	
	if (actual_encoding < mime_binary)
	{
		long ll = length;
		unsigned char const *p;
		
		for (p = first; p < end; ++p)
		{
			int ch = *p;
			unsigned ae = char_encode[ch];
			if (ch == '\r')
			{
				++p; /* must be lf, since not binary */
				if (actual_max_line_length < ll)
					actual_max_line_length = ll;
				ll = 0;
			}
			else
			{
				++ll;
			}
			if (actual_encoding < ae)
				actual_encoding = ae;
		}
		
		length = ll;
		ment->actual_max_line_length = actual_max_line_length;
		ment->actual_encoding = actual_encoding;
	}
	
	return length;
}

static long ment_offset(MIME_ENTITY *const ment, FILE *fp, mime_error* merro)
/*
** get file offset and possibly start new entity with
** a suitably aligned offset
*/
{
	long offset /* compiler happy */ = 0;
	mime_error me = MIME_NO_ERROR;
	
	if (fp == NULL ||
		ferror(fp) ||
		fseek(fp, 0, SEEK_END) != 0 ||
		(offset = ftell(fp)) < 0)
			me = MIME_SYSFILE_ERROR;
	
	else if (ment->is_leaf_part)
	{
		if (ment->length == 0) /* new entity */
		{
			while ((offset & 0xf) != 0) /* align at 32bit */
			{
				if (fputc(0, fp) == EOF)
				{
					me = MIME_SYSFILE_ERROR;
					break;
				}
				++offset;
			}
			if (me == MIME_NO_ERROR)
				ment->offset = offset;
		}
		else if (ment->offset + ment->length < offset)
			me = MIME_FILE_ERROR;
	}
	else
		me = MIME_FILE_ERROR;

	if (me == MIME_NO_ERROR)
		return offset; /* good return */
	
	/* error return */
	if (merro)
		*merro = me;
	return -1;
}

static int read_single_body(MIME* const m, int discard)
/*
** assume buffer at start of single-body
** return cur boundary level (>= 0) if terminated by boundary,
** or -1 if terminated by error, -2 by eof
*/
{
#if defined(TEST_MAIN) && !defined(NDEBUG)
	int qp_escape_hack_start_counter = 0;
	int qp_escape_hack_found_counter = 0;
#endif		
	
	/* eventually will see CRLF + boundary + "--" */
	/* not used: size_t const b_size_max = max_boundary_size(m) + 2; */
	
	size_t line_length = 0;
	MIME_ENTITY* const ment = discard ? NULL : m->ment + m->ment_cur;
	int const encoding = discard ? mime_binary : ment->encoding;
	int rtc = -1;

	assert(mime_invariant(m));
	assert(discard == 0 || discard == 1);
	assert(discard || ment->offset == 0);
	assert(discard || ment->length == 0);
	/* not used: assert(b_size_max + 2 -- for CRLF -- < m->buf_max); */
	
	if (!discard)
		ment_offset(ment, m->fpw, NULL); /* was: ment->offset = ftell(m->fpw); */

	for (;;) /* outer refill-buffer loop */
	{
		static unsigned const base64_shift[] = {18, 12, 6, 0};
		int ndx64 = 0;
		unsigned long ulb64 = 0;
		int b64_discard = 0;
		unsigned char sbuf[4];
		int qp_escape = 0; /* -ale: 21sep2002 these definition moved here */
		
		size_t in_buffer = fill_buffer(m);
		unsigned char * p = m->buf_first, *t = p;
		unsigned char const * const e = m->buf_end;
		unsigned char *end_of_boundary = NULL;

		while (p < e) /* inner scan-buffer loop */
		{
			int const ch = *p;

			assert(t <= p);

			if (ch == '\r' || ch == '\n')
			{
				/*
				** make sure enough of a new line is in buffer
				** (if not, ch will be read again next loop)
				*/
				assert(MAX_LINE_LENGTH + MAX_LINE_LENGTH < MAX_BUFFERED_AHEAD);
				if (p + MAX_LINE_LENGTH + MAX_LINE_LENGTH > e &&
					in_buffer >= m->buf_max)
				{
					if (ndx64 > 0 && b64_discard == 0)
					/*
					** cannot handle line breaking anywhere, should be at 4th char
					** -ale: 21sep2002 added check for b64_discard
					*/
					{
						assert(discard == 0);
						ment->error_base64 = 1;
						/* -ale: 21sep2002: omitted: ndx64 = ulb64 = 0; */
					}
					goto end_of_buffer;
				}

				/*
				** if not "\n", must be "\r\n" to be a line break
				*/
				if (ch == '\n' || p + 1 < e && p[1] == '\n')
				{
					++p;
					if (ch == '\r')
						++p;

					/*
					** check for known (non-empty) boundaries
					*/
					if (p + 2 < e && p[0] == '-' && p[1] == '-')
					{
						unsigned avail = e - (p + 2);
						rtc = check_boundary_candidate(m, p + 2, &avail, check_min);
						if (rtc >= 0) /* found boundary */
						{
							end_of_boundary = p + avail + 2;
							
							/* leave line break in buffer */
							--p;
							if (ch == '\r')
								--p;

							if (!discard)
								ment->found_term_boundary = 1;
							goto end_of_buffer;
						}
					}
					
					if (qp_escape == 1) /* soft line break: ignore it */
					{
						assert(encoding == mime_quoted_printable);
						qp_escape = 0;
					}
					else if (encoding == mime_base_64)
						; /* line breaks in b64 is just whitespace: ignore it */
					else if (encoding != mime_binary && !discard)
					/*
					** we write "\r\n" for text. for quoted printable, we may
					** have room in the buffer (t < p). If t == p, write "\r",
					** save the buffer and save "\n", then refill the buffer.
					*/
					{
						assert(qp_escape == 0 || qp_escape == 2);
						if (qp_escape)
						{
							ment->error_quoted_printable = 1;
							qp_escape = 0;
						}
						
						*t++ = '\r';
						if (ch == '\n' && t >= p)
						{
							qp_escape = '\r';
#if defined(TEST_MAIN) && !defined(NDEBUG)
							++qp_escape_hack_start_counter;
#endif		
							goto end_of_buffer;
						}
						*t++ = '\n';
					}
					else /* binary or discard */
					{
						++t;
						if (ch == '\r')
							++t;
						assert(t == p);
					}
					assert(t <= p);
				}
				else /* spare '\r' */
				{
					if (!discard)
						ment->actual_encoding = mime_binary;
					switch (encoding)
					{
						case mime_base_64:
						case mime_quoted_printable:
							/* silently discard white space */
							++p;
							break;

						default:
							/* accept, since not unencoding */
							++p;
							++t;
							assert(p == t);
							break;
					}
				}
			}
			else if (!discard) /* not line break */
			{
				++p;
				
				/*
				** decode char
				*/
				switch (encoding)
				{
					case mime_base_64:
					{
						unsigned long ul;
						if ((unsigned)ch < sizeof base64_tab &&
							(ul = base64_tab[ch]) < 64)
						{
							ulb64 |= ul << base64_shift[ndx64++];
							if (ndx64 >= 4)
							{
								*t++ = (ulb64 >> 16) & 0xff;
								*t++ = (ulb64 >> 8) & 0xff;
								*t++ = ulb64 & 0xff;
								ulb64 = ndx64 = 0;
								assert(t <= p);
							}
						}
						else if (ch == '=')
							++b64_discard;
						else if (!isspace(ch))
							ment->error_base64 = 1;
						break;
					}
					case mime_quoted_printable:
						if (qp_escape > 0)
						{
							sbuf[qp_escape - 1] = ch;
							if (qp_escape == 1)
							{
								qp_escape = 2;
							}
							else
							{
								char *term;
								unsigned long ul;
								sbuf[2] = 0;
								ul = strtoul(sbuf, &term, 16);
								if (term == (char*)&sbuf[2])
									*t++ = ul;
								else
								{
									*t++ = '=';
									*t++ = sbuf[0];
									*t++ = sbuf[1];
									ment->error_quoted_printable = 1;
									assert(t <= p);
								}
								assert(qp_escape == 2);
								qp_escape = 0;
							}
						}
						else if (ch == '=')
						{
							if (p + 2 >= e)
							{
								if (in_buffer >= m->buf_max)
								/* avoid decoding at the end of buffer */
								{
									--p;
									goto end_of_buffer;
								}
								
								/* "=" is the last or last but one char in file */
								ment->error_quoted_printable = 1;
								*t++ = ch;
								assert(t <= p);
							}
							else
								qp_escape = 1;
						}
						else
						{
							*t++ = ch;
							assert(t <= p);
						}
						break;
						
					default: /* 7bit, 8bit, binary or undefined */
						++t;
						assert(t == p);
						break;
				}
			}
			else /* discard */
			{
				assert(t == p);
				t = ++p;
			}
		}

		/* break from inner loops w/o consuming chars */
		end_of_buffer:

		assert(t <= p);
		assert(t >= m->buf_first);
		
		if (!discard)
		{
			/*
			** write unencoded data to temporary file
			*/
			size_t w = t - m->buf_first;
			if (w > 0 && w != fwrite(m->buf_first, 1, w, m->fpw))
			{
				if (ment->length == 0)
					mime_clear_content(m);
				ment->error_system = 1;
				return -1;
			}
			if (qp_escape == '\r')
			{
				putc('\n', m->fpw);
				++ment->length;
				qp_escape = 0;
#if defined(TEST_MAIN) && !defined(NDEBUG)
				++qp_escape_hack_found_counter;
				assert(qp_escape_hack_start_counter ==
					qp_escape_hack_found_counter);
#endif		
			}
			
			ment->length += w;
			line_length = sniff_data(m, line_length, m->buf_first, t);
			/* -ale: 21sep2002 the switch was only done on end-of data
			** -- was: if (rtc >= 0 || in_buffer <= 0)
			** -ale: 28sep2002 the switch now was not called on end-of file
			** -- was if (p < e)
			** just do the switch every time to adjust decodings
			*/
			switch (encoding)
			{
				case mime_base_64:
					assert(ndx64 < 4);
					if (ndx64 > 0)
					{
						if (ndx64 + b64_discard != 4)
						{
							b64_discard = 4 - ndx64;
							ment->error_base64 = 1;
						}
						if (b64_discard >= 3)
						/* -ale: 19feb2002: is illegal */
						{
							b64_discard = 2;
							ment->error_base64 = 1;
						}
						
						sbuf[0] = (ulb64 >> 16) & 0xff;
						/* -ale: 19feb2002: omitted: if (b64_discard <= 1) */
							sbuf[1] = (ulb64 >> 8) & 0xff;
						/* -ale: 19feb2002: will ever be written
						if (b64_discard <= 2)
							sbuf[2] = ulb64 & 0xff;  */
						w = 3 - b64_discard;
						ment->length += fwrite(sbuf, 1, w, m->fpw);
						sniff_data(m, line_length, sbuf, sbuf + w);
					}
					break;
					
				case mime_quoted_printable:
					if (qp_escape > 1)
						ment->error_quoted_printable = 1;
					/* -ale: useless: qp_escape = 0; */
					break;
	
				default:
					break;
			}
		}
		
		m->buf_first = p;

		if (!discard && (ferror(m->fpw) || ferror(m->fp)))
		{
			if (m->last_error != MIME_INFILE_ERROR)
				m->last_error = MIME_SYSFILE_ERROR;
			ment->error_system = 1;
			rtc = -1;
			break;
		}
		if (rtc >= 0)       /* found boundary */
		{
			assert(end_of_boundary);
			m->buf_first = end_of_boundary;
			break;
		}
		if (in_buffer <= 0) /* found eof */
		{
			rtc = -2;
			break;
		}
	}

#if defined(TEST_MAIN) && !defined(NDEBUG)
	//if (!be_silent)
	fprintf(stderr, "%sment=%d, rtc=%d, hack=%d,%d%s\n",
		discard ? "DISCARD " : "",
		m->ment_cur, rtc,
		qp_escape_hack_start_counter,
		qp_escape_hack_found_counter,
		qp_escape_hack_start_counter == qp_escape_hack_found_counter ? "" :
			"WRONG COUNT OF BUFFER HACKs!!!!!!!!!!!");
#endif

	return rtc;
}

static int add_entity(MIME* const m, unsigned cur)
/* append an entity at given position; ment_cur and flags not changed */
{
	unsigned ndx = m->ment_count;
	
	assert(mime_invariant(m));
	assert(cur <= m->ment_count);
	
	if (++m->ment_count >= m->ment_max)
	{
		size_t const ment_max = 2 * m->ment_max;
		MIME_ENTITY* ment;
		
		if (ment_max > m->ment_max && /* overflow (??) */
			ment_max < SIZE_MAX / sizeof(MIME_ENTITY))
		{
			ment = (MIME_ENTITY*)malloc(ment_max * sizeof(MIME_ENTITY));
			if (ment == NULL)
				m->last_error = MIME_SYSTEM_ERROR;
		}
		else
		{
			ment = NULL;
			m->last_error = MIME_OVERFLOW_ERROR;
		}
		
		if (ment == NULL)
		{
			--m->ment_count;
			return -1;
		}

		memcpy(ment, m->ment, m->ment_max * sizeof(MIME_ENTITY));
		free(m->ment);
		m->ment = ment;
		m->ment_max = ment_max;
	}
	
	if (ndx > cur)
	{
		MIME_ENTITY *const ment = m->ment + cur;
		size_t count = ndx - cur;
		
		memmove(ment + 1, ment, count * sizeof(MIME_ENTITY));
	}
	mime_entity_init(m->ment + cur, m->next_internal_id++);
	
	/* not true mime_invariant(m): must set flags */
	return 0;
}

static int read_body_part(MIME*, int, int); /* forward decl for recursive call */

static int read_multipart_body(MIME* const m)
/*
** assume buffer at start of body
** return cur boundary level (>= 0) if terminated by boundary
** or -1 on error, -2 on eof
*/
{
	int const cur = m->ment_cur;
	unsigned char const *const boundary = m->ment[cur].boundary;
	size_t const b_size = boundary ? strlen(boundary) : 0;
	int rtc, found_boundary = 0;

	assert(mime_invariant(m));
	assert(m->ment[cur].boundary != NULL);
	assert(b_size);
	assert(b_size < MAX_BOUNDARY_LENGTH); /* 19dec2001 */
	/* assert(b_size + 4 < MAX_BUFFERED_AHEAD); is implied by previous */
	/* -ale: 16feb2004: not enforced: assert(m->ment[cur].is_discrete == 0); */

	/*
	** check for boundary, discard any preamble
	*/
	rtc = 0;

#if 0 // -ale: 16feb2004 boundary consumed	
	while (check_buffer(m, b_size + 4)) 
	{
		unsigned char *p = m->buf_first;
		if (p[0] == '-' && p[1] == '-' && memcmp(p+2, boundary, b_size) == 0)
		{
			p += b_size + 2;
			m->buf_first = p;
			rtc = discard_upto_crlf(m);
			break;
		}
		
		if (discard_upto_crlf(m))
			break;
	}
#endif
	
	while (boundary_may_fit_in_buffer(m))
	{
		unsigned char *p = m->buf_first;
		unsigned char const * const e = m->buf_end;
		if (p[0] == '-' && p[1] == '-')
		{
			unsigned avail = e - (p + 2);
			if (avail >= b_size && memcmp(p+2, boundary, b_size) == 0)
			{
				p += b_size + 2;
				m->buf_first = p;
				rtc = discard_upto_crlf(m);
				found_boundary = 1;
				break;
			}
			else if ((rtc =
				check_boundary_candidate(m, p + 2, &avail, check_min)) >= 0)
			{
				m->buf_first = p + 2 + avail;
				break;
			}
		}

		if (discard_upto_crlf(m))
			break;
	}

	if (!found_boundary)
	/*
	** if it has no starting boundary, it is not really
	** a multipart entity. e.g. headers of an rfc822 entity.
	*/
	{
		MIME_ENTITY * const ment = m->ment + cur - 1;
		m->ment[cur].is_discrete = 1;
		
		if (cur > 0 &&
			ment->is_leaf_part == 0 &&
			ment->mime_subtype == mime_message_rfc822)
				ment->mime_subtype = mime_message_rfc822_headers;
		else
			m->ment[cur].error_structure = 1;
	}

	else if (rtc == 0)
	/*
	** add any number of ment[cur]'s children
	*/
	{
		size_t last_cur /* compiler happy */ = 0;
		int is_first_part = 1;
		
		for (;;)
		{
			unsigned char *p;
			
			if (add_entity(m, m->ment_cur + 1))
				return -1;

			if (is_first_part)
			{
				m->ment[cur].is_leaf_part = 0;
				is_first_part = 0;
			}
			else
				m->ment[last_cur].is_last_part = 0;

			last_cur = ++m->ment_cur;

			m->ment[last_cur].is_last_part = 1;
			m->ment[last_cur].is_leaf_part = 1;

			rtc = read_body_part(m, 1, 0);
			if (rtc != cur)
			{
				if (rtc > cur)
					rtc = -1;
				/*
				** if 0 <= rtc < cur then my boundary is missing
				** and an ancestor boundary has been found
				*/
				break;
			}

			/*
			** expect "--" for end of part (epilogue),
			** or ['\r'] '\n' for another part
			*/
			if (!check_buffer(m, 1))
			{
				rtc = -2; // end of file
				break;
			}

			p = m->buf_first;
			if (p[0] == '-' &&
				check_buffer(m, 2) && m->buf_first[1] == '-')
			{
				int ertc;

				m->buf_first += 2;

				/*
				** discard epilogue
				*/
				if ((ertc = read_single_body(m, 1)) >= cur)
					rtc = -1;
				else
					rtc = ertc;

				break;
			}
			else if (p[0] == '\r' &&
				check_buffer(m, 2) && m->buf_first[1] == '\n')
			{
				m->buf_first += 2;
			}
			else if (p[0] == '\n')
			{
				m->buf_first += 1;
			}
			else
			{
				rtc = -1;
				break;
			}
		}
	}
	
	if (rtc == -1 && m->last_error == 0)
		m->last_error = MIME_DATA_ERROR;

	assert(mime_invariant(m));
	return rtc;
}

static int read_body_part(MIME* const m, int read_headers, int const save_headers)
/*
** current mime entity must have been initialized already
** return cur boundary level (>= 0) or -1 on error, -2 on eof
*/
{
	int rtc = 0;
	int const cur = m->ment_cur;
	MIME_ENTITY * const ment = m->ment + cur;
	char const *p;
	
	assert(mime_invariant(m));
	/* if saving headers must also read them */
	assert(save_headers == 0 || read_headers != 0);
	
	ment->found_term_boundary = 0; // still reading it
	if (read_headers)
		while ((rtc = read_mime_header(m, save_headers)) > 0)
			continue;

	if (ment->encoding == 0)
		ment->encoding = m->default_encoding;
	if (ment->charset == 0)
		ment->charset = m->default_charset;

	if (!ment->is_leaf_part)
		rtc = -1;

	/*
	** missing or too long boundary on multipart
	*/
	if (rtc == 0 && ment->is_discrete == 0 &&
		ment->mime_subtype != mime_message_rfc822 &&
			(ment->boundary == NULL ||
			strlen(ment->boundary) > MAX_BUFFERED_AHEAD))
	{
		if (ment->error_system)
			rtc = -1;
		else
		{
			if (ment->boundary)
				ment->boundary[MAX_BUFFERED_AHEAD] = 0;
			else /* missing boundary */
			{
				ment->is_discrete = 1;
				ment->error_structure = 1;
				ment->mime_type = mime_application;
				ment->mime_subtype = mime_application_octetstream;
			}
		}
	}
	
	/*
	** boundary found in the headers (missing two empty lines)
	** for empty body return pretending it has been read
	*/
	if (rtc == 0 && ment->found_term_boundary)
	{
		rtc = m->ment_cur;
		m->ment_cur = cur;
		ment->warning_structure = 1;
		
		if (rtc < cur)
		{
			assert(mime_invariant(m));
			return rtc;
		}

		/*
		** was the first boundary in this multipart?
		*/
		ment->found_term_boundary = 0;
		rtc = rtc == cur && ment->is_discrete == 0 ? 0 : -1;
	}

	/*
	** missing one crlf before boundary as in:
   ** --myboundary
   ** Content-Type: application/x-empty
   ** X-note: missing newline before boundary, after end of headers
   ** 
   ** --myboundary
   ** ^-- we are here
	** (e.g. IE4.5 Mac for multipart/form-data on type 'file'?)
	*/
	if (rtc == 0 &&
		ment->is_discrete && /* -ale: bugfix: 28jan02: added this line */
		check_buffer(m, 2) &&
		*(p = m->buf_first) == '-' && p[1] == '-')
	{
		char const * const e = find_end_of_line(m);

		if (e && e > p + 2)
		{
			unsigned avail = e - (p + 2);
			int const ie_rtc =
				check_boundary_candidate(m, p + 2, &avail, check_min);
			if (ie_rtc >= 0)
			{
				m->buf_first += 2 + avail;
				ment->found_term_boundary = 1;
				assert(mime_invariant(m));
				return ie_rtc;
			}
		}
	}
	
	/*
	** read entity body
	*/
	if (rtc == 0)
	{		
		if (ment->mime_subtype == mime_message_rfc822)
		{
			size_t last_cur;
			
			if (add_entity(m, m->ment_cur + 1))
				return -1;
	
			last_cur = ++m->ment_cur;
			m->ment[cur].is_leaf_part = 0; /* ment != &m->ment[cur], possibly */
			m->ment[last_cur].is_last_part = 1;
			m->ment[last_cur].is_leaf_part = 1;
			
			rtc = read_body_part(m, 1, 1);
		}
		else if (ment->mime_type == mime_multipart)
			rtc = read_multipart_body(m);
		else /* was: if (ment->is_discrete) */
			rtc = read_single_body(m, 0);
	}

	if (rtc == -1 && m->last_error == 0)
		m->last_error = MIME_DATA_ERROR;

	assert(mime_invariant(m));
	return rtc;
}

int mime_read(MIME *m, FILE *fp)
/*
** read from m->fp, and write to mime binary file (m->fpw).
** fp input is expected to be at body, i.e. either a single-body or
** a multipart-body, as indicated by m.
**
** caller should initialize the current entry by passing in any
** relevant header, then call here with fp positioned at the body;
** fp will be read until eof, discarding any prologue/epilogue
** (including the last epilogue that goes through end-of-file).
**
** return non-zero for any error
** for warnings, return 0 but set a warning status indicator
** call mime_missing_term_boundary to test termination
*/
{
	int rtc;
	size_t cur = m->ment_cur, i;
	assert(mime_invariant(m));
	
	/*
	** cannot read into a node / NULL arg
	*/
	if (!m->ment[cur].is_leaf_part || fp == NULL)
	{
		m->last_error = MIME_CALL_ERROR;
		return -1;
	}

	/*
	** set and check input and output files
	*/
	m->fp = fp;
	if (m->fpw == NULL ||
		ferror(m->fpw) || ferror(fp) ||
		fseek(m->fpw, 0, SEEK_END) != 0)
	{
		m->last_error = MIME_SYSFILE_ERROR;
		return -1;
	}

	/*
	** reset buffer before read; this discards any
	** content previously kept there via mime_get_content
	*/
	m->buf_first = m->buf_end = m->buffer;
	m->ment_committed = -1;
	m->unix_line_break = 0;
	m->header_u_count = m->header_b_count = 0;

	rtc = read_body_part(m, 0, 0);
	
#if defined(TEST_MAIN)
	//if (!be_silent)
	fprintf(stderr, "count=%d, discrete=%d, rtc=%d from read_body_part\n",
		m->ment_count, m->ment->is_discrete, rtc);
#endif /* TEST_MAIN */	

	if (rtc != -2)
		rtc = -1;
	else
		rtc = 0;

	/*
	** check each multipart has children and no error structure
	*/
	for (i = cur; i <= m->ment_cur && rtc == 0; ++i)
	{
		MIME_ENTITY const*const ment = m->ment + i;
		if (ment->is_discrete != ment->is_leaf_part ||
			ment->error_structure || ment->error_system)
				rtc = -1;
	}
	
	/*
	** reset buffer, position, input file pointer
	*/
	m->buf_first = m->buffer;
	m->buf_end = m->buffer + m->buf_max; /* temp storage */
	m->ment_cur = cur;
	m->fp = NULL;
	
	assert(rtc == -1 || mime_invariant(m)); /* not sure... */
	
	if (rtc == -1 && m->last_error == 0)
		m->last_error = MIME_DATA_ERROR;
	return rtc;
}

int mime_missing_term_boundary(MIME const * const m)
{
	int rlev = 0,  rtc = 0;
	size_t cur = m->ment_cur, i;
	assert(mime_invariant(m));

	if (m->ment[cur].is_leaf_part)
		return 0;
	
	for (i = cur + 1; cur < m->ment_count && rlev >= 0; ++i)
	{
		MIME_ENTITY const*const ment = m->ment + i;
		if (ment->is_leaf_part)
		{
			if (!ment->found_term_boundary)
				++rtc;
			if (ment->is_last_part)
				--rlev;
		}
		else
			++rlev;
	}

	return rtc;
}

/* -----------------------------------------------------
** MIME EDIT
*/

size_t mime_get_count_all(MIME const * const m)
{
	assert(mime_invariant(m));
	return m->ment_count;	
}

int mime_is_leaf_part(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->is_leaf_part;
}

int mime_is_empty(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->is_leaf_part && ment->length == 0;
}

char const * mime_get_name(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->name;
}

char const * mime_get_filename(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->filename;
}

char const * mime_get_content_type(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->content_type;
}

enum mime_type_enum mime_get_mime_type(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->mime_type;
}

enum mime_subtype_enum mime_get_mime_subtype(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->mime_subtype;
}

char const * mime_get_boundary(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->boundary;
}

char const * mime_get_description(MIME const * const m)
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	return ment->description;
}

enum mime_encoding_enum mime_get_encoding(MIME const * const m)
	{ assert(mime_invariant(m));
	return m->ment[m->ment_cur].encoding; }

enum mime_charset_enum mime_get_charset(MIME const * const m)
	{ assert(mime_invariant(m));
	return m->ment[m->ment_cur].charset; }

enum mime_disposition_enum mime_get_disposition(MIME const * const m)
	{ assert(mime_invariant(m));
	return m->ment[m->ment_cur].dispo; }

static int set_string(MIME *const m, unsigned char **d, char const *s)
{
	char *p;

	if (s)
	{
		if ((p = strdup(s)) == NULL)
		{
			m->last_error = MIME_SYSTEM_ERROR;
			return -1;
		}
	}
	else
		p = NULL;

	if (*d)
		free(*d);
	*d = p;

	return 0;
}

/* these replace the old value and return 0 or
** -1 if strdup failed or bad content_type.
** use mime_set_header for setting unknown values
*/
int mime_set_boundary(MIME *m, char const *s)
	{ assert(mime_invariant(m));
	return set_string(m, &m->ment[m->ment_cur].boundary, s); }
int mime_set_description(MIME *m, char const *s)
	{ assert(mime_invariant(m));
	return set_string(m, &m->ment[m->ment_cur].description, s); }
	
int mime_set_content_type(MIME *m, char const *h)
{
	int rtc;
	size_t const cur = m->ment_cur;
	MIME_ENTITY* const ment = m->ment + cur;
	mime_header_func const fn = mime_field_name[mime_content_type].fn;
	
	assert(mime_invariant(m));
	
	rtc = mime_set_header_internal(ment, h, fn);
	if (rtc < 0)
		m->last_error = MIME_HEADER_ERROR;
	else if (rtc >= 0)
		m->ment_committed = -1;
	if (ment->error_system)
		m->last_error = MIME_SYSTEM_ERROR;

	assert(mime_invariant(m));
	if (rtc == 1) /* accepted */
		return 0; /* ok */
	return -1;
}

int mime_set_name(MIME *m, char const *s)
	{ assert(mime_invariant(m));
	return set_string(m, &m->ment[m->ment_cur].name, s); }
int mime_set_filename(MIME *m, char const *s)
	{ assert(mime_invariant(m));
	return set_string(m, &m->ment[m->ment_cur].filename, s); }

enum mime_encoding_enum
mime_set_encoding(MIME *m, enum mime_encoding_enum en)
	{ assert(mime_invariant(m));
	return m->ment[m->ment_cur].encoding = en; }

enum mime_charset_enum
mime_set_charset(MIME *m, enum mime_charset_enum ch)
	{ assert(mime_invariant(m));
	return m->ment[m->ment_cur].charset = ch; }

enum mime_disposition_enum
mime_set_disposition(MIME *m, enum mime_disposition_enum dispo)
	{ assert(mime_invariant(m));
	return m->ment[m->ment_cur].dispo = dispo; }

int mime_unmap(MIME*m)
{
	MIME_ENTITY *const ment = m->ment + m->ment_cur;
	
	assert(mime_invariant(m));

	return mime_entity_unmap(ment);
}

int mime_mmap_content(MIME*m, content_ptr*result)
/*
** mmap the binary representation of the entity and fill the resulting
** pointer area, or just retrieve the current pointer area
** return 0 if successful
*/
{
	MIME_ENTITY *const ment = m->ment + m->ment_cur;
	mmap_area *const mmarea = &ment->mmarea;
	
	assert(mime_invariant(m));

	if (ment->is_discrete && ment->length >= 0)
	{
		if (mmarea->flag == mmap_area_empty)
		{
			if ((mmarea->length = ment->length) >= m->sys_pagesize)
			{
				FILE *const fp = ment->is_alt_file ? ment->fpw : m->fpw;
				off_t offset =  ment->offset, spare = offset % m->sys_pagesize;
				
				fflush(fp);
				offset -= spare;
				mmarea->length += spare;
				if ((mmarea->start = mmap(NULL, mmarea->length,
					PROT_READ | PROT_WRITE, MAP_SHARED, fileno(fp), offset)) != MAP_FAILED)
				{
					mmarea->flag = mmap_area_mapped;
					mmarea->adjusted.start = mmarea->start + spare;
					mmarea->adjusted.length = mmarea->length - spare;
				}
			}
			else if ((mmarea->start = malloc(mmarea->length)) != NULL)
			{
				if ((long)mime_read_content(m,
					mmarea->start, mmarea->length, 0) != mmarea->length)
				{
					free(mmarea->start);
					mmarea->start = NULL;
				}
				else
					mmarea->flag = mmap_area_malloced;
				mmarea->adjusted.start = mmarea->start;
				mmarea->adjusted.length = mmarea->length;
			}
			else
				errno = ENOMEM;
		}
		
		if (mmarea->flag != mmap_area_empty)
		{
			*result = mmarea->adjusted;
			return 0;
		}
		m->last_error = MIME_SYSTEM_ERROR;
	}
	else // ?? if (m->last_error < MIME_CALL_ERROR)
		m->last_error = MIME_CALL_ERROR;
	return -1;
}

size_t mime_read_content(MIME*m, void*buf, size_t bufmax, long offset)
/*
** read up to bufmax bytes from offset in buffer and return bytes read.
** offset is relative to the beginning of the content.
** if less than bufmax bytes are read, it is because the content is over.
** in case of error return -1, so bufmax must be less than that
*/
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;

	assert(mime_invariant(m));

	if (ment->is_discrete && ment->length >= 0)
	{
		FILE *const fp = ment->is_alt_file ? ment->fpw : m->fpw;
		size_t size = bufmax;
		long const left = ment->length - offset;

		if (left <= 0 || size <= 0)
			return 0;
		if ((unsigned long)left <= SIZE_MAX && (size_t)left < size)
			size = (size_t) left;
		if (fp != NULL &&
			fseek(fp, ment->offset + offset, SEEK_SET) == 0 &&
			fread(buf, 1, size, fp) == size)
				return size;
	}
	
	return -1;
}

char const *mime_get_content(MIME*m, int keep)
/*
** read the whole content of an entity into the buffer
** if keep is nonzero, the content will be kept into the
** buffer until next mime_read, mime_clear_buffer or mime_clear.
**
** if the entity is not discrete or its content is too long
** will return NULL;
*/
{
	MIME_ENTITY const*const ment = m->ment + m->ment_cur;

	assert(mime_invariant(m));

	if (ment->is_discrete && ment->length < INT_MAX && ment->length >= 0)
	{
		size_t size = (size_t) ment->length;
		unsigned char *const p = m->buf_first;
		FILE *fp = ment->is_alt_file ? ment->fpw : m->fpw;
		if (fp != NULL &&
			size + 1 < mime_get_available_buffer(m) &&
			(size <= 0 ||
				fseek(fp, ment->offset, SEEK_SET) == 0 &&
				fread(p, 1, size, fp) == size))
		{
			p[size] = 0;
			if (keep)
				m->buf_first += size + 1;
			return p;
		}
	}
	
	return NULL;
}

unsigned long mime_get_content_length(MIME const * const m)
{
	assert(mime_invariant(m));
	return m->ment[m->ment_cur].length;
}

size_t mime_get_available_buffer(MIME const * const m)
{
	assert(mime_invariant(m));
	return m->buf_max - (size_t)(m->buf_first - m->buffer);
}

int mime_running_out_buffer(MIME const * const m)
{
	size_t avail = mime_get_available_buffer(m);
	assert(mime_invariant(m));
	
	return avail < 1024 || avail < m->buf_max/4;
}
	
void *mime_detach_buffer(MIME *const m)
/*
** allocate a new buffer and pass back the old one
** on error, return NULL
*/
{
	char *new_buf = (char*)malloc(m->buf_max), *old_buf = m->buffer;
	
	assert(mime_invariant(m));
	if (new_buf == NULL)
		return NULL;
	
	m->buffer = new_buf;
	m->buf_first = m->buffer;
	m->buf_end = m->buffer + m->buf_max; /* temp storage */
	
	assert(mime_invariant(m));
	return old_buf;
}


int mime_clear_buffer(MIME* const m)
{
	assert(mime_invariant(m));
	m->buf_first = m->buffer;
	return 0;
}

int mime_clear_boundary_all(MIME*m)
{
	size_t i;
	size_t const cur = m->ment_cur;
	
	assert(mime_invariant(m));
	
	for (i = 0; i < m->ment_count; ++i)
		if (m->ment[i].boundary)
		{
			m->ment_cur = i;
			mime_set_boundary(m, NULL);
		}
	m->ment_cur = cur;

	assert(mime_invariant(m));
	return 0;
}

int mime_clear_encoding_all(MIME*m)
{
	size_t i;
	
	assert(mime_invariant(m));
	
	for (i = 0; i < m->ment_count; ++i)
		m->ment[i].encoding = 0;
		
	m->ment_committed = -1;
	assert(mime_invariant(m));
	return 0;
}

int mime_clear_content(MIME* const m)
/*
** discard the content of the current discrete entity
** by resetting its length and return 0;
** if entity is not leaf return -1
*/
{
	MIME_ENTITY *const ment = m->ment + m->ment_cur;
	int rtc = -1;
	
	assert(mime_invariant(m));

	if (ment->is_leaf_part)
	{
		ment->offset = 0;
		ment->length = 0;
		ment->actual_max_line_length = 0;
		ment->actual_encoding = 0;
		ment->is_discrete = 1;
		ment->is_alt_file = 0;
		ment->error_base64 = 0;
		ment->error_quoted_printable = 0;
		ment->error_system = 0;
		if (ment->mmarea.flag != mmap_area_empty)
			rtc = mime_entity_unmap(ment);
		else
			rtc = 0;
	}
	m->ment_committed = -1;
	
	assert(mime_invariant(m));
	return rtc;
}

static void clear_boundary(MIME* const m,
	unsigned char const *boundary, unsigned char const *end_boundary)
/* delete any duplicate boundary and check unique_boundary */
{
	size_t cur = m->ment_cur, match;
	unsigned avail = end_boundary - boundary;
	m->ment_cur = m->ment_count - 1;
	
	while ((int)(match = 
		check_boundary_candidate(m, boundary, &avail, check_max)) >= 0)
	{
		m->ment_cur = match;
		mime_set_boundary(m, NULL);
		m->ment_cur = m->ment_count - 1;
	}
	m->ment_cur = cur;
}

int mime_add_line_of_content(MIME* const m, char const *fmt, ...)
/*
** append a line on a discrete entity and return number of bytes written,
** including the end of line (crlf) which is appended after the line.
** entity content must be empty or else the entity must be the last
** one written to. the formatted line must fit in the available buffer
** and contain no CR and no LF.
** 0 is returned for error.
*/
{
	size_t cur = m->ment_cur;
	MIME_ENTITY *const ment = m->ment + cur;
	FILE *outfp = ment->is_alt_file ? ment->fpw : m->fpw;
	unsigned char *const p = m->buf_first;
	size_t const avail = mime_get_available_buffer(m);

	assert(mime_invariant(m));
	assert(fmt != NULL);

	if (ment_offset(ment, outfp, &m->last_error) != -1)
	{
		va_list stack;
		size_t written, w;
		
		va_start(stack, fmt);
		written = vsnprintf(p, avail, fmt, stack);
		va_end(stack);
		if ((int)written > 0 && written + 1 < avail &&
			memchr(p, '\n', written) == NULL &&
			memchr(p, '\r', written) == NULL)
		{
			if (ment->actual_max_line_length < (long)written)
				ment->actual_max_line_length = written;
				
			/* check for boundaries in text and invalidate anyone found */
			if (written > 2 && p[0] == '-' && p[1] == '-')
				clear_boundary(m, p + 2, p + written);

			w = fwrite(p, 1, written, outfp);
			sniff_data(m, 0, p, p + written);
			w += length_fputs("\r\n", outfp);
			if (w == written + 2)
			{
				ment->length += w;
				m->ment_committed = -1;
				ment->is_discrete = 1;
				return w;
			}
			
			m->last_error = MIME_SYSFILE_ERROR;
			if (ment->length == 0)
				mime_clear_content(m);
			else
				ment->error_system = 1;
		}
		else
			m->last_error = MIME_BUFFER_ERROR;
	}
	return 0;
}

int mime_add_text_of_content(MIME* const m, char const *text)
/*
** append 0 terminated text, one line at a time.
** a terminating newline is appended if not already present.
** return 0 if ok
** -1 is returned for error.
*/
{
	size_t cur = m->ment_cur;
	MIME_ENTITY *const ment = m->ment + cur;
	FILE *outfp = ment->is_alt_file ? ment->fpw : m->fpw;
	int rtc = -1;

	assert(mime_invariant(m));
	assert(text != NULL);

	if (ment_offset(ment, outfp, &m->last_error) != -1)
	{
		size_t written = 0;
		char const *p1 = text, *const endp = text + strlen(text);
		
		rtc = 0;		
		while (p1 < endp)
		{
			size_t w1, w2;
			char const *p2 = strchr(p1, '\n');
			if (p2 == NULL)
				p2 = endp;
			w1 = p2 - p1;
			if (ment->actual_max_line_length < (long)w1)
				ment->actual_max_line_length = w1;
			if (w1 > 2 && p1[0] == '-' && p1[1] == '-')
				clear_boundary(m, p1 + 2, p2);
			
			w2 = fwrite(p1, 1, w1, outfp);
			sniff_data(m, 0, p1, p2);
			w2 += length_fputs("\r\n", outfp);
			if (w2 != w1 + 2)
			{
				m->last_error = MIME_SYSFILE_ERROR;
				if (ment->length == 0)
					mime_clear_content(m);
				else
					ment->error_system = 1;
				rtc = -1;
				break;
			}
			written += w2;
			p1 = p2 + 1;
		}
		
		if (rtc == 0)
			ment->length += written;
	}
	return 0;
}

int mime_add_file_of_content(MIME* const m, FILE* const fp, int is_text)
/*
** append each line, with CR+LF on a discrete entity.
** entity content must be empty or else the entity must be the last
** one written to.
** 0 is returned, or -1 for error.
** the buffer must contain 1024 free bytes at time of call.
*/
{
	size_t cur = m->ment_cur;
	MIME_ENTITY *const ment = m->ment + cur;
	FILE *outfp = ment->is_alt_file ? ment->fpw : m->fpw;
	int rtc = -1;
	unsigned char *const p = m->buf_first;
	size_t avail = mime_get_available_buffer(m);

	assert(mime_invariant(m));
	assert(fp);

	if (avail < 1024)
		m->last_error = MIME_BUFFER_ERROR;
	else if (ment_offset(ment, outfp, &m->last_error) != -1)
	{
		size_t line = 0;
		char *s = fgets(p, avail, fp);
		int error_system = ment->error_system;

		--avail; /* one for line break (if adding \r) */
		
		while (s && !error_system)
		{
			size_t w = strlen(s);
			if (w <= 0 || w >= avail)
			{
				error_system = 1;
				break;
			}
			if (is_text &&
				s[w-1] == '\n' &&
				s[w-2] != '\r')
			{
				s[w-1] = '\r';
				s[w++] = '\n';
			}
			if (ment->actual_encoding < mime_binary)
			{
				if (line == 0 && w > 2 && s[0] == '-' && s[1] == '-')
					clear_boundary(m, s + 2, s + w);
				line = sniff_data(m, line, s, s + w);
			}
			if (w != fwrite(s, 1, w, outfp))
			{
				error_system = 1;
				break;
			}
			ment->length += w;
			s = fgets(p, avail, fp);
		}
		m->ment_committed = -1;
		ment->is_discrete = 1;
		ment->found_term_boundary = 1;
		if (ferror(fp))
			error_system = 1;
		if (error_system)
		{
			ment->error_system = 1;
			m->last_error = MIME_SYSFILE_ERROR;
			if (ment->length == 0)
				mime_clear_content(m);
		}
		else
		{
			ment->error_system = 0;
			rtc = 0;
		}
	}
	return rtc;
}

int mime_attach_binary(MIME* const m, FILE *fp_alt)
/*
** fp_alt is opened and closed by caller, it must remain
** open until mime_clear_content is called on this entity
** or until mime_clear.
** fp_alt must be a regular seekable file. it need not be
** writable if add_line_of_content won't be called for this
** function.
** the entity must be an empty leaf part.
** return 0 if ok.
*/
{
	size_t cur = m->ment_cur;
	MIME_ENTITY *const ment = m->ment + cur;

	assert(mime_invariant(m));
	assert(fp_alt);

	if (!mime_is_empty(m))
		m->last_error = MIME_CALL_ERROR;
	else if (fseek(fp_alt, 0, SEEK_END) != 0 ||
		(ment->length = ftell(fp_alt)) == -1)
			m->last_error = MIME_SYSFILE_ERROR;
	else
	{
		ment->actual_encoding = mime_binary;
		ment->is_discrete = 1;
		ment->is_alt_file = 1;
		ment->found_term_boundary = 1;
		ment->fpw = fp_alt;
		m->ment_committed = -1;
		return 0;
	}
	return -1;
}

int mime_child_of(MIME const * const m, size_t target)
{
	assert(mime_invariant(m));
	return target <= m->ment_cur &&
		m->ment_cur <= seek_last_child(m, target);
}

size_t mime_get_depth(MIME const * const m)
{
	assert(mime_invariant(m));

	return get_level(m, m->ment_cur);
}

size_t mime_get_position(MIME const*const m)
{
	assert(mime_invariant(m));
	return m->ment_cur;
}

int mime_set_position(MIME*const m, size_t cur)
{
	assert(mime_invariant(m));
	if (cur >= m->ment_count)
	{
		m->last_error = MIME_CALL_ERROR;
		return -1;
	}
	m->ment_cur = cur;
	return 0;	
}

int mime_get_internal_id(MIME const*const m)
{
	assert(mime_invariant(m));
	return m->ment[m->ment_cur].internal_id;
}

int mime_search_internal_id(MIME *const m, int internal_id)
/*
** set current position to the entity with the required id
** return 0 if successful, -1 if no such id
*/
{
	size_t i;
	assert(mime_invariant(m));
	
	for (i = 0; i < m->ment_count; ++i)
		if (m->ment[i].internal_id == internal_id)
		{
			m->ment_cur = i;
			assert(mime_invariant(m));
			return 0;
		}
	
	return -1;
}

int mime_seek(MIME *m, mime_seek_arg arg)
/*
** move the current position in the tree as directed by parameters
** and return the index of (new) current position. If the position
** the last call to mime_seek caused no change of position, a call
** to mime_seek_is_stuck will return non-zero.
*/
{
	size_t cur = m->ment_cur;
	assert(mime_invariant(m));

	switch (arg)
	{
		case mime_seek_top:
			cur = 0;
			break;
			
		case mime_seek_downl:
			if (!m->ment[cur].is_leaf_part)
				++cur;
			break;
			
		case mime_seek_downr:
			if (!m->ment[cur].is_leaf_part)
				cur = seek_last_sibling(m, cur + 1);
			break;
			
		case mime_seek_right:
			if (!m->ment[cur].is_last_part)
			{
				if (!m->ment[cur].is_leaf_part)
					cur = seek_last_child(m, cur + 1);
				++cur;
			}
			break;
		
		case mime_seek_left:
		case mime_seek_up:
			if (cur > 0)
				left_or_up(m, 0, cur, &cur, arg);
			break;

		case mime_seek_first:
			if (cur > 0)
				left_or_up(m, 0, cur, &cur, mime_seek_up);
			++cur;
			break;

		case mime_seek_last:
			cur = seek_last_sibling(m, cur);
			break;
			
		case mime_seek_next:
			if (cur + 1 < m->ment_count)
				++cur;
			break;

		case mime_seek_prev:
			if (cur > 0)
				--cur;
			break;

		default:
			break;
	}
	
	if (cur >= m->ment_count)
		cur = m->ment_cur;
	
	if (cur == m->ment_cur)
		m->seek_stuck = 1;
	else
	{
		m->seek_stuck = 0;
		m->ment_cur = cur;
	}

	assert(mime_invariant(m));
	return cur;
}

int mime_seek_is_stuck(MIME const * const m)
{
	assert(mime_invariant(m));
	return m->seek_stuck;
}

int mime_set_header(MIME* const m, mime_header_id hid, char const *h)
/*
** if hid is mime_header_unknown, h must be of the form "name: value",
** else hid identifies the header name and h must be in the form "value".
** return 0 if header ok but not recognized, 1 if ok, -1 if error
*/
{
	int rtc;
	size_t cur = m->ment_cur;
	MIME_ENTITY* const ment = m->ment + cur;
	unsigned char const * const old_boundary = ment->boundary;
	mime_header_func fn;
	
	assert(mime_invariant(m));
	
	if ((int)hid < 0 || hid >= mime_header_unknown)
		fn = NULL;
	else
		fn = mime_field_name[hid].fn;

	rtc = mime_set_header_internal(ment, h, fn);
	if (rtc < 0)
		m->last_error = MIME_HEADER_ERROR;
	else if (rtc == 1 && ment->boundary && old_boundary != ment->boundary)
	{
		size_t i;
		
		/* on added boundary, check it is different (why?) */
		for (i = 0; i < m->ment_count; ++i)
			if (i != cur && m->ment[i].boundary &&
				strcmp(ment->boundary, m->ment[i].boundary) == 0)
				{
#if defined(TEST_MAIN)
					if (!be_silent)
					fprintf(stderr, "At %d freeing boundary %d (%s)\n",
						cur, i, m->ment[i].boundary);
#endif
					free(m->ment[i].boundary);
					m->ment[i].boundary = NULL;
				}
	}
	if (ment->error_system)
		m->last_error = MIME_SYSTEM_ERROR;
	m->ment_committed = -1;

	assert(mime_invariant(m));
	return rtc;
}

int mime_mark_for_discard(MIME*m)
{
	MIME_ENTITY *const ment = m->ment + m->ment_cur;
	assert(mime_invariant(m));
	
	if (ment->is_discrete)
	{
		ment->is_for_discard = 1;
		return 0;
	}
	m->last_error = MIME_CALL_ERROR;
	return -1;
}

static size_t commit_discard_beyond(MIME* const m, size_t const cur)
{
	size_t i, first_to_del, last_to_del, removed_last, last_to_keep;
	size_t count = 0, to_keep = 0, to_del = 0;
	int const target_level = get_level(m, cur) + 1;
	
	assert(mime_invariant(m));
	assert(cur < m->ment_count);
	assert(!m->ment[cur].is_leaf_part);
	assert(m->ment_cur == 0);
	
	first_to_del = last_to_del = last_to_keep = 0; /* compiler happy */
	
	/*
	** count the number of children to delete/keep
	** recursively delete any marked part below child nodes
	*/
	for (i = cur + 1; i < m->ment_count;)
	{
		int const level = get_level(m, i);
		
		if (level == target_level)
		{
			if (m->ment[i].is_for_discard)
			{
				if (to_del++ == 0)
					first_to_del = last_to_del = i;
				else
					last_to_del = i;
			}
			else
			{
				if (!m->ment[i].is_leaf_part)
				{
					size_t n = commit_discard_beyond(m, i);
					if (n > 0) /* start over again (??) */
					{
						i = cur + 1;
						to_keep = to_del = 0;
						count += n;
						assert(target_level == get_level(m, cur) + 1);
						continue; 
					}
				}
				++to_keep;
				last_to_keep = i;
			}
		}
		else if (level < target_level)
			break; /* don't do nephews */
		++i;
	}
	
	/*
	** nothing to discard ?
	*/
	if (to_del == 0)
		return count;
	
	/*
	** mark cur for discard if all nodes below it are gone
	*/
	if (to_keep == 0)
	{
		unsigned is_last = m->ment[cur].is_last_part;
		mime_entity_clear(&m->ment[cur]);
		mime_entity_init(&m->ment[cur], m->next_internal_id++);
		m->ment[cur].is_for_discard = 1;
		m->ment[cur].is_last_part = is_last;
		m->ment[cur].is_leaf_part = 1;
	}

	/*
	** do remove any number of consecutive nodes
	*/
	removed_last = 0;
	while (first_to_del <= last_to_del)
	{
		size_t n;
		for (i = first_to_del; i <= last_to_del; ++i)
		{
			if (!m->ment[i].is_for_discard)
				break;
			removed_last |= m->ment[i].is_last_part;
			mime_entity_clear(&m->ment[i]);
		}
		
		/* remove [first, i) */
		n = i - first_to_del;
		assert(i > first_to_del);
		assert(to_del >= n);
		assert(i <= m->ment_count);

		if (i < m->ment_count) /* if not end of array */
			memmove(m->ment + first_to_del, m->ment + i, 
				(m->ment_count - i) * sizeof(MIME_ENTITY));
		last_to_del -= n;
		if (last_to_keep > first_to_del)
			last_to_keep -= n;
		to_del -= n;
		m->ment_count -= n;
		count += n;
		
		/* now old-i-th entity is in first_to_del */
		assert(first_to_del > last_to_del || !m->ment[first_to_del].is_for_discard);
		
		for (i = first_to_del + 1; i <= last_to_del; ++i)
		{
			int level = get_level(m, i);
		
			if (level == target_level && m->ment[i].is_for_discard)
				break;
		}
		
		first_to_del = i;
	}
	
	if (removed_last != 0 && to_keep > 0)
		m->ment[last_to_keep].is_last_part = 1;

	assert(to_del == 0);
	assert(mime_invariant(m));
	return count;
}

int mime_commit_discard(MIME* const m)
/*
** return number of elements deleted
*/
{
	int count;
	assert(mime_invariant(m));

	m->ment_cur = 0; /* leave current on top */
	if (m->ment[0].is_leaf_part)
		count = 0;
	else
		count = commit_discard_beyond(m, 0);

	/*
	** on deleting last element, convert to empty text
	*/
	if (m->ment[0].is_for_discard)
	{
		assert(m->ment_count == 1);
		
		mime_entity_clear(m->ment);
		mime_entity_init(m->ment, m->next_internal_id++);
		m->ment->is_last_part = 1;
		m->ment->is_leaf_part = 1;
	}

	m->ment_committed = -1;

	assert(mime_invariant(m));
	return count;	
}

int mime_wrap(MIME* const m)
/*
** wrap a multipart/mixed entity around current entity
** new entity becomes current. return nonzero on error
*/
{
	int rtc;
	size_t cur = m->ment_cur;
	
	assert(mime_invariant(m));

	if ((rtc = add_entity(m, cur)) == 0)
	{
		MIME_ENTITY *ment = &m->ment[cur];
		
		mime_init_multipart_mixed(ment, m->next_internal_id++);
		
		ment[0].is_last_part = ment[1].is_last_part;
		/* already memset: ment[0].is_leaf_part = 0; */
		ment[1].is_last_part = 1; /* lone part */
	}
	
	m->ment_committed = -1;

	assert(mime_invariant(m));
	return rtc;
}

static int mime_insert_internal(MIME* const m, size_t cur)
/*
** insert new entity at position cur, shifting right
** whatever is there. return nonzero on error
*/
{
	int rtc;

	assert(mime_invariant(m));
	assert(m->ment_count > 1); /* must wrap */
	assert(cur <= m->ment_count);
		
	if ((rtc = add_entity(m, cur)) == 0)
	{
		MIME_ENTITY *ment = &m->ment[cur];

		ment->is_leaf_part = 1;
		/* already memset: ment->is_last_part = 0; */
	}
	
	/* not true after append: assert(mime_invariant(m)); */
	return rtc;
}

int mime_insert(MIME*m)
{
	size_t cur = m->ment_cur;
	int rtc = -1;

	assert(mime_invariant(m));
	
	if (cur == 0) /* must wrap */
		m->last_error = MIME_CALL_ERROR;
	else if (mime_insert_internal(m, cur) == 0)
	{
		rtc = 0;
		m->ment_committed = -1;
	}
	
	assert(mime_invariant(m));
	assert(m->ment_cur == cur);
	return rtc;
}

int mime_append(MIME*m)
{
	int level;
	size_t cur = m->ment_cur, i;
	assert(mime_invariant(m));
	
	if (cur == 0)
	{
		m->last_error = MIME_CALL_ERROR;
		return -1; /* must wrap */
	}
	mime_seek(m, mime_seek_right);
	if (!mime_seek_is_stuck(m))
		return mime_insert(m);
	
	assert(m->ment[cur].is_last_part != 0);
	
	level = get_level(m, cur);
	for (i = cur + 1; i < m->ment_count; ++i)
		if (get_level(m, i) < level)
			break;
	
	if (mime_insert_internal(m, i) != 0)
		return -1;
	
	m->ment[i].is_last_part = 1;
	m->ment[cur].is_last_part = 0;
	m->ment_cur = i;

	m->ment_committed = -1;

	assert(mime_invariant(m));
	assert(get_level(m, i) == level);
	return 0;		
}

/* -----------------------------------------------------
** MIME WRITE
*/

static int write_saved_headers(MIME* const m)
/*
** read back the header data and send it to the output file
** return -1 for error, 1 if end-of-line was the last char output,
** 0 otherwise
*/
{
	MIME_ENTITY * const ment = m->ment + m->ment_cur;
	FILE *infp = ment->is_alt_file ? ment->fpw : m->fpw;
	FILE *outfp = m->fp;
	int newline = 1;	
	long total = ment->hlength;
	int const dot_stuffing = m->dot_stuffing;

	assert(mime_invariant(m));
	assert(ment->hlength < LONG_MAX); /* what else??? */

	if (infp == NULL || ferror(infp) ||
		 outfp == NULL || ferror(outfp) ||
		 fseek(infp, ment->hoffset, SEEK_SET) != 0)
	{
		if (infp == NULL || outfp == NULL)
			m->last_error = MIME_FILE_ERROR;
		else
			m->last_error = MIME_SYSFILE_ERROR;
		return -1;
	}
	
	while (total > 0 && !ferror(infp) && !ferror(outfp))
	{
		unsigned char buf[1024];
		size_t size = total >= (long)(sizeof buf) ?
			sizeof buf : (size_t)total;
		unsigned char *p = buf;
		unsigned char const * const ebuf = p + size;
		
		if (fread(buf, 1, size, infp) != size)
			break;

		total -= size;
		while (p < ebuf && !ferror(outfp))
		{
			int ch = *p++;
			if (dot_stuffing && ch == '.' && newline)
			{
				fputs("..", outfp); /* duplicate dot */
				newline = 0;
			}
			else if ((newline = ch == '\n') != 0)
				fputs(m->crlf, outfp);
			else if (ch != '\r')
				putc(ch, outfp);
		}
	}
		
	if (ferror(infp) || ferror(outfp))
		m->last_error = MIME_SYSFILE_ERROR;
	else if (total != 0)
		m->last_error = MIME_FILE_ERROR;
	else
		return newline; /* good return */

	/* failure */
	return -1;
}

static int write_single_body(MIME* const m)
/*
** read back the body data and send it encoded to the output file
** return -1 for error, 1 if end-of-line was the last char output,
** 0 otherwise
*/
{
	MIME_ENTITY * const ment = m->ment + m->ment_cur;
	FILE *infp = ment->is_alt_file ? ment->fpw : m->fpw;
	FILE *outfp = m->fp;
	unsigned encoding = ment->encoding;
	int newline = 1;
	int const is_text = ment->encoding < mime_binary &&
		ment->mime_type == mime_text;
	int const dot_stuffing = m->dot_stuffing;

	
	static unsigned const base64_shift[] = {16, 8, 0};
	size_t linelength = 0, ndx64 = 0;
	unsigned long ulb64 = 0;
	
	int binary_cr = 0;
	long total = ment->length;

	assert(mime_invariant(m));
	assert(ment->length < LONG_MAX); /* what else??? */

	if (infp == NULL || ferror(infp) ||
		 outfp == NULL || ferror(outfp) ||
		 fseek(infp, ment->offset, SEEK_SET) != 0)
	{
		if (infp == NULL || outfp == NULL)
			m->last_error = MIME_FILE_ERROR;
		else
			m->last_error = MIME_SYSFILE_ERROR;
		return -1;
	}
	
	while (total > 0 && !ferror(infp) && !ferror(outfp))
	{
		unsigned char buf[1024];
		size_t size = total >= (long)(sizeof buf) ?
			sizeof buf : (size_t)total;
		unsigned char *p = buf;
		unsigned char const * const ebuf = p + size;
		
		if (fread(buf, 1, size, infp) != size)
			break;

		total -= size;
		switch (encoding)
		{
			case mime_base_64:
				while (p < ebuf && !ferror(outfp))
				{
					unsigned long ch = *p++;
					ulb64 |= ch << base64_shift[ndx64++];
					if (ndx64 >= 3)
					{
						if (linelength + 4 > MAX_LINE_LENGTH)
						{
							fputs(m->crlf, outfp);
							linelength = 0;
						}
						putc(bin2base64_tab[(ulb64 >> 18) & 0x3f], outfp);
						putc(bin2base64_tab[(ulb64 >> 12) & 0x3f], outfp);
						putc(bin2base64_tab[(ulb64 >> 6) & 0x3f], outfp);
						putc(bin2base64_tab[ulb64 & 0x3f], outfp);
						linelength += 4;
						ulb64 = ndx64 = 0;
					}
				}
				break;					
			
			case mime_quoted_printable:
				while (p < ebuf && !ferror(outfp))
				{
					int ch = *p++;
					int quoted_len = 1;
					if (isalnum(ch))
					{
						if (newline && ch == 'F' &&
							(p + 3 >= ebuf || 
								isspace(p[4]) && strincmp(p, "rom", 3) == 0))
									quoted_len = 3;
						else
							quoted_len = 1;
					}
					else if (ch == '\r')
					{
						if (is_text)
							continue;
						quoted_len = 3;
					}
					else if (ch >= 127 || iscntrl(ch) || ch == '=')
						quoted_len = 3;
#if 0 /* -ale: change on 30apr03: no need to set spaces in text */
					else if ((ch == ' ' || ch == '\t') &&
						(p >= ebuf || isspace(*p)))
							quoted_len = 3;
#endif							
					else if (newline && ch == '.')
						quoted_len = 3;

					if (ch == '\n' && is_text)
					{
						fputs(m->crlf, outfp);
						newline = 1;
						linelength = 0;
					}
					else
					{
						if (linelength + quoted_len > MAX_LINE_LENGTH)
						{
							fprintf(outfp, "=%s", m->crlf);
							linelength = 0;
							if (ch == '.') /* -ale: fix on 1jun00 */
								quoted_len = 3;
						}
						
						if (quoted_len == 3)
							fprintf(outfp, "=%02X", ch);
						else
							putc(ch, outfp);
						linelength += quoted_len;
						newline = 0;
					}
				}
				break;
			
			case mime_7bit:
			case mime_8bit:
				while (p < ebuf && !ferror(outfp))
				{
					int ch = *p++;
					if (dot_stuffing && ch == '.' && newline)
					{
						fputs("..", outfp); /* duplicate dot */
						newline = 0;
					}
					else if ((newline = ch == '\n') != 0)
						fputs(m->crlf, outfp);
					else if (ch != '\r')
						putc(ch, outfp);
				}
				break;
			
			case mime_binary:
			/*
			** dot stuffing and unix line breaks not considered,
			** except keeping track of any trailing newline for
			** a final line break on text data (see mime_write)
			*/
				fwrite(buf, size, 1, outfp);
				if (m->crlf[0] == '\r')
				{
					if (size == 1)
					{
						if (binary_cr)
							newline = buf[0] == '\n';
						else
						{
							newline = 0;
							binary_cr = buf[0] == '\r';
						}
					}
					else
						newline = buf[size - 2] == '\r' &&
							buf[size - 1] == '\n';
				}
				else
					newline = buf[size - 1] == '\n';
				break;
				
			default:
				break;
		}
	}
	
	if (encoding == mime_base_64)
	{
		if (ndx64 > 0)
		{
			if (linelength + 4 > MAX_LINE_LENGTH)
				fputs(m->crlf, outfp);
			putc(bin2base64_tab[(ulb64 >> 18) & 0x3f], outfp);
			putc(bin2base64_tab[(ulb64 >> 12) & 0x3f], outfp);
			putc(ndx64 == 1 ? '=' : bin2base64_tab[(ulb64 >> 6) & 0x3f], outfp);
			putc('=', outfp);
		}
		fputs(m->crlf, outfp);
		newline = 1;
	}
	
	if (ferror(infp) || ferror(outfp))
		m->last_error = MIME_SYSFILE_ERROR;
	else if (total != 0)
		m->last_error = MIME_FILE_ERROR;
	else
		return newline; /* good return */

	/* failure */
	return -1;
}

static MIME_ENTITY*
write_preamble(MIME* const m, int level, MIME_ENTITY *ment, size_t part1)
/*
** expect newline already written
*/
{
	FILE* const outfp = m->fp;
	assert(mime_invariant(m));
	
	if (level == 1) /* root level */
		fprintf(outfp,
			" This is a multipart MIME message%s"
			" level %10s%s",
				m->crlf, "part", m->crlf);

	for (;; ++ment)
	{
		unsigned is_last = ment->is_last_part;
		set_content_type_string(ment);
		if (ment->error_system)
		{
			m->last_error = MIME_SYSTEM_ERROR;
			return NULL;
		}
			
		fprintf(outfp, "%6d %.*s Part#%u: %s%s",
			level, level % 10, "..........", (size_t)(ment - m->ment) - part1 + 1,
				ment->content_type, m->crlf);
			
		if (!ment->is_leaf_part)
			ment = write_preamble(m, level + 1, ment + 1, part1);
		if (is_last || ment == NULL)
			return ment;
	}
}

static int write_body_part(MIME*, int); /* forward decl for recursive call */

static int write_multipart_body(MIME* const m)
/*
** assume headers have been written and terminated with an empty line
** write any number of sub-entities and return 1 if the last output
** was an end-of-line sequence, -1 on error, 0 otherwise
*/
{
	const unsigned cur = m->ment_cur;
	MIME_ENTITY * ment = m->ment + cur;
	char const * const boundary = ment->boundary;
	FILE *outfp = m->fp;
	int newline = 1;

	assert(mime_invariant(m));

	/* should commit encoding before write */
	if (ment->boundary == NULL || ment->is_discrete != 0 || outfp == NULL)
	{
		m->last_error = MIME_CALL_ERROR;
		return -1;
	}

	/*
	** write preamble only at first level
	*/
	if (cur == m->ment_committed)
	{
		if (write_preamble(m, 1, ment, cur) == NULL)
			return -1;
		fputs(m->crlf, outfp);
	}

	/*
	** write first boundary and first body-part
	*/
	fprintf(outfp, "--%s%s", boundary, m->crlf);
	++m->ment_cur;
	newline = write_body_part(m, 0);
	
	/*
	** write any more body-parts, until last
	*/
	while (newline >= 0 && !(++ment)->is_last_part)
	{
		fprintf(outfp, "%s--%s%s", m->crlf, boundary, m->crlf);
		++m->ment_cur;
		newline = write_body_part(m, 0);
	}
	
	/*
	** write end boundary + empty line as epilogue + CRLF
	*/
	if (newline >= 0)
	{
		fprintf(outfp, "%s--%s--%s%s", m->crlf, boundary, m->crlf, m->crlf);
		newline = 1;
	}

	assert(mime_invariant(m));
	return newline;
}

static int write_body_part(MIME* const m, int add_headers)
/*
** current mime entity must have been initialized already
** return -1 for error, 1 if end-of-line was the last char output,
** 0 otherwise
*/
{
	int rtc = 1;
	size_t const cur = m->ment_cur;
	MIME_ENTITY * const ment = m->ment + cur;
	unsigned from_headers = cur == m->ment_committed ? 0 : 1; /* from 0 if first body */
	
	assert(mime_invariant(m));

   /*
   ** write headers, possibly including saved headers
   */
   if (add_headers)
   {
   	from_headers = 0;
   	if (ment->hlength > 0)
   		rtc = write_saved_headers(m);
   }
   mime_write_header_internal(ment, m->fp, m->crlf, from_headers);
   fputs(m->crlf, m->fp);

	/*
	** write entity body if no error
	*/
	if (rtc < 0 || ment->error_system != 0)
	{
		m->last_error = MIME_SYSTEM_ERROR;
		rtc = -1;
	}
	else if (ment->is_discrete)
		rtc = write_single_body(m);
	else if (ment->mime_subtype == mime_message_rfc822 ||
		ment->mime_subtype == mime_message_rfc822_headers)
	{
		++m->ment_cur;
		rtc = write_body_part(m, 1);
	}
	else
		rtc = write_multipart_body(m);

	assert(mime_invariant(m));
	return rtc;
}

int mime_write(MIME *m, FILE *fp, int unix_line_break, int dot_stuffing)
/*
** assume non-mime headers have been written already and
** mime_commit_content has been called successfully.
**
** mime headers for the current entity are written to fp,
** followed by an empty line and the entity body
** (including any sub entity)
*/
{
	int rtc = -1;
	size_t cur = m->ment_cur;
	assert(mime_invariant(m));

	m->fp = fp;
	if (unix_line_break)
		strcpy(m->crlf, "\n");
	else
		strcpy(m->crlf, "\r\n");
	m->dot_stuffing = dot_stuffing ? 1 : 0;
	
	if (fp == NULL || m->ment_committed != cur)
		m->last_error = MIME_CALL_ERROR;
	else if (ferror(fp))
		m->last_error = MIME_SYSFILE_ERROR;
	else
	{
		rtc = write_body_part(m, 1);
	
		/*
		** make sure there is a trailing line break
		** for [e]smtp transports or text single bodies
		*/	
		if (rtc == 0 &&
			(m->ment[cur].encoding < mime_binary ||
				m->ment[cur].mime_type == mime_text))
					fputs(m->crlf, fp);

		if (rtc >= 0)
			rtc = ferror(fp);
	}
	
	m->ment_cur = cur;
	assert(mime_invariant(m));
	if (ferror(fp) && m->last_error == 0)
		m->last_error = MIME_SYSFILE_ERROR;
	return rtc;
}

int mime_extract_single_body(MIME* m, FILE *fp,
	enum mime_encoding_enum en, int unix_line_break)
/*
** extract the content of a discrete entity, using the encoding
** and the line break desired, write it to the file at current
** position and return 0 if ok
*/
{
	int rtc = -1;
	size_t cur = m->ment_cur;
	
	assert(mime_invariant(m));

	m->fp = fp;
	if (unix_line_break)
		strcpy(m->crlf, "\n");
	else
		strcpy(m->crlf, "\r\n");
	
	if (fp == NULL ||
		!m->ment[cur].is_discrete ||
		!m->ment[cur].is_leaf_part ||
		en <= 0 || en > mime_binary)
			m->last_error = MIME_CALL_ERROR;
	else if (ferror(fp))
		m->last_error = MIME_SYSFILE_ERROR;
	else
	{
		enum mime_encoding_enum old_en = m->ment[cur].encoding;
		m->ment[cur].encoding = en;
		rtc = write_single_body(m);
		m->ment[cur].encoding = old_en;
		
		/*
		** make sure there is a trailing line break
		** for [e]smtp transport or text
		*/	
		if (rtc == 0 &&
			(en < mime_binary ||
				m->ment[cur].mime_type == mime_text))
					fputs(m->crlf, fp);
					
		if (rtc >= 0)
			rtc = ferror(fp);
	}
	
	m->ment_cur = cur;
	assert(mime_invariant(m));
	if (ferror(fp))
		m->last_error = MIME_SYSFILE_ERROR;
	return rtc;
}

static int commit_encoding_below(MIME* const m, size_t i,	
	enum mime_encoding_enum en, mime_encoding_rtc *result)
{
	assert(mime_invariant(m));
	assert(result != NULL);
	assert(i < m->ment_count);
			
	for (;; ++i)
	{
		MIME_ENTITY* const ment = m->ment + i;
		
		if (ment->is_leaf_part != ment->is_discrete &&
			*result < mime_bad_structure)
				*result = mime_bad_structure;
		
		set_content_type_string(ment);
		if (ment->error_system)
			*result = mime_system_error;
			
		if (ment->is_leaf_part)
		{
			enum mime_encoding_enum as_is = ment->encoding;
			switch (as_is) /* see if can keep encoding as_is */
			{
				case mime_base_64:
				case mime_quoted_printable:
				/* keep as_is */
					break;
					
				case mime_7bit:
				case mime_8bit:
					if (ment->actual_max_line_length > 998 ||
						(enum mime_encoding_enum)(ment->actual_encoding) > as_is ||
						as_is == mime_8bit &&
							ment->actual_encoding <= mime_7bit) /* change */
								as_is = mime_encoding_not_set;
					else if (as_is == mime_8bit &&
						*result < mime_ok_8bit) /* keep as_is */
							*result = mime_ok_8bit;
					break;

				case mime_binary:
					if (ment->mime_type == mime_text) /* change */
						as_is = mime_encoding_not_set;
					else if (*result < mime_ok_binary) /* keep as_is */
						*result = mime_ok_binary;
					break;
					
				case mime_encoding_not_set:
				default:
					as_is = mime_encoding_not_set; /* change */
					break;
			}
			
			if (as_is == mime_encoding_not_set)
				switch (en) /* set new encoding */
				{
					case mime_base_64:
					case mime_quoted_printable:
						/*
						** set as requested
						*/
						ment->encoding = en;
						/* if (...) *result = mime_ok_7bit */
						break;
						
					case mime_7bit: /* smtp */
					case mime_8bit: /* esmtp */
						if (ment->actual_max_line_length > 998 ||
							(enum mime_encoding_enum)(ment->actual_encoding) > en)
						{
							/*
							** cannot encode long lines or 8bit as requested
							*/
							if (ment->mime_type == mime_text &&
								ment->actual_encoding != mime_binary)
									ment->encoding = mime_quoted_printable;
							else
								ment->encoding = mime_base_64;
						}
						else if (en == mime_8bit &&
							ment->actual_encoding <= mime_7bit)
							/*
							** useless to require 8bit if 7bit is enough
							*/
							ment->encoding = mime_7bit;
						else
							/*
							** do as requested
							*/
							ment->encoding = en;
						
						if (ment->encoding == mime_8bit &&
							*result < mime_ok_8bit)
								*result = mime_ok_8bit;
						break;
						
					case mime_binary:
						if (ment->mime_type == mime_text &&
							ment->actual_encoding < mime_binary)
						{
							/*
							** leave 7/8 bit text as such
							*/
							ment->encoding = ment->actual_encoding;
							if (ment->encoding < mime_7bit)
								ment->encoding = mime_7bit;
							else if (ment->encoding == mime_8bit &&
								*result < mime_ok_8bit)
									*result = mime_ok_8bit;
						}
						else /* set binary as requested */
						{
							ment->encoding = mime_binary;
							if (*result < mime_ok_binary)
								*result = mime_ok_binary;
						}
						break;
						
					case mime_encoding_not_set:
					default:
						if (*result < mime_bad_argument)
							*result = mime_bad_argument;
						break;
				}
			ment->found_term_boundary = 1;
		}
		else /* set boundary and encoding according to children */
		{
			mime_encoding_rtc oldresult;

			if (ment->boundary == NULL)
			{
				char buf[128];
				/* format arguments:
				** 1) boundary_rand, 2) any number, 3) boundary_unique
				*/
				sprintf(buf, boundary_template,
					m->boundary_rand, i - m->ment_cur + 1, m->boundary_unique);
				if ((ment->boundary = strdup(buf)) == NULL)
					*result = mime_system_error;
			}

			oldresult = *result;
			*result = mime_ok_7bit;
			i = commit_encoding_below(m, i + 1, en, result);
			if (*result < mime_bad_result)
				ment->encoding = *result + mime_identity;
			if (*result < oldresult)
				*result = oldresult;
			
		}
		
		if (ment->is_last_part)
			return i;
	}
}

mime_encoding_rtc
mime_commit_encoding(MIME* const m, enum mime_encoding_enum en)
/*
** to be called before mime_write, passing the transport capability
** (i.e. mime_binary for http transport, mime_8bit for 8BITMIME capable
** esmtp, and mime_7bit for other smtp transport.) Lower encodings can
** also be selected.
**
** The function is meant to be called before mime_write in order to
** make sure there are no errors before sending a file and get the
** value (7BIT / 8BITMIME) for the BODY parameter of MAIL FROM
** (e.g. MAIL FROM:<some@one> BODY=8BITMIME) for esmtp transport.
**
** For each node, select the encoding as appropriate, taking into
** account (i) the transport capability, (ii) the actual content of
** the data, (iii) the current encoding, and (iv) for multiparts,
** the encoding of any inner part.
**
** The caller may set desired encodings after committing them if
** specific conversions are desired (e.g. base64 vs. quoted-printable)
**
** return self explanatory enumerated value
*/
{
	mime_encoding_rtc result = mime_ok_7bit;
	
	assert(mime_invariant(m));
	
	if (commit_boundary_unique(m) != 0)
		result = mime_cannot_do_boundary;
	commit_encoding_below(m, m->ment_cur, en, &result);
	if (result < mime_bad_result)
		m->ment_committed = m->ment_cur;
	else switch(result)
	{
		case mime_bad_argument:
		case mime_bad_structure:
			m->last_error = MIME_CALL_ERROR;
			break;
		case mime_system_error:
			m->last_error = MIME_SYSTEM_ERROR;
			break;
		default:
			m->last_error = MIME_INTERNAL_ERROR;
			break;
	}
	assert(mime_invariant(m));
	return result;
}


/*******************************************************************/
#if defined(TEST_MAIN)
enum fn_type { tv, tv_v, ti_m, ti_mi, ti_mseek, tcc_m, tcc_mi, ti_mcc, ti_mf_report,
	ti_mI, ti_mIi, ti_mfii, ti_mfeu, te_me, ti_me};
enum fn_parm { no_parm, any_int_parm, any_string_parm,
	output_file_parm, input_file_parm, encoding_parm, seek_parm,
	unix_break_parm, dot_stuffing_parm};

typedef struct debug_command_tag
{
	union mime_fn_sig
	{
		void *v;
		int (*i_m)(MIME*);
		int (*i_mi)(MIME*, int);
		char const *(*cc_m)(MIME*);
		char const *(*cc_mi)(MIME*, int);
		int (*i_mcc)(MIME*, char const *);
		int (*i_mf_report)(MIME*, FILE*);
		int (*i_mI)(MIME*, FILE*);
		int (*i_mIi)(MIME*, FILE*, int);
		int (*i_mfii)(MIME*, FILE*, int, int);
		int (*i_mfeu)(MIME*, FILE*, enum mime_encoding_enum, int);
		int (*e_me)(MIME*, enum mime_encoding_enum);
		void (*v_v)(void);
	} fn;
	char *name;
	enum fn_type fn_t;
	char const *desc;
	char const *key;
	int parm[8];
} debug_command;

typedef struct file_param_array_tag
{
	size_t cur;
	FILE *fp[16];
} file_param_array;

static int mime_testread(MIME* m, FILE* fp);

void debug_fn_help(void);
int report_mime(MIME *m, FILE *fp, size_t i, int indent, int all);
void report_mime_struct(MIME *m, FILE *fp);
int report_mime_all(MIME *m, FILE *fp)
	{ return report_mime(m, fp, 0, 0, 1); }
int report_mime_one(MIME *m, FILE *fp)
	{ return report_mime(m, fp, m->ment_cur, 0, 0); }
void leak_memory(void)
	{ malloc(8); }

int line_of_content(MIME* m, char const *s)
{
	int w = mime_add_line_of_content(m, "%s", Psz(s));
	if (w == 0)
		w = -1;
	return w;
}

int test_set_header(MIME* const m, char const *h)
{
	return mime_set_header(m, mime_header_unknown, h);
}

debug_command dbgtab[] =
{
	{{ mime_testread}, "-", ti_mI, "read whole mail file into current entity", "rdf",
								{input_file_parm}},
	{{ mime_read}, "mime_read", ti_mI, "read mime body into current entity", "read",
								{input_file_parm}},
	{{ mime_missing_term_boundary}, "mime_missing", ti_m, "check missing boundaries", "miss",
								{no_parm}},
	{{ mime_get_name}, "mime_get_name", tcc_m, "get name", "gn", {no_parm}},
	{{ mime_get_filename}, "mime_get_filename", tcc_m, "get filename", "gf", {no_parm}},
	{{ mime_get_content_type}, "mime_get_content_type", tcc_m, "get content type", "gf",
								{no_parm}},
	{{ mime_get_boundary}, "mime_get_filename", tcc_m, "get boundary", "gb", {no_parm}},
	{{ mime_get_description}, "mime_get_filename", tcc_m, "get description", "gd",
								{no_parm}},
	{{ mime_get_encoding}, "mime_get_encoding", ti_m,
								"get Content-Transfer-Encoding", "ge",
								{no_parm}},
	{{ mime_get_charset}, "mime_get_charset", ti_m, "get charset", "gchar",
								{no_parm}},
	{{ mime_get_disposition}, "mime_get_disposition", ti_m,
								"get Content-Disposition", "gdsipo",
								{no_parm}},
	{{ mime_set_name}, "mime_set_name", ti_mcc, "set name", "hname",
								{any_string_parm}},
	{{ mime_set_filename}, "mime_set_filename", ti_mcc, "set filename", "hfile",
								{any_string_parm}},
	{{ mime_set_content_type}, "mime_set_content_type", ti_mcc, "set Content-Type string",
								"hcont",
								{any_string_parm}},
	{{ mime_set_boundary}, "mime_set_boundary", ti_mcc, "set boundary", "hboundary",
								{any_string_parm}},
	{{ mime_set_description}, "mime_set_description", ti_mcc,
								"set description", "hdesc",
								{any_string_parm}},
	{{ mime_set_encoding}, "mime_set_encoding", ti_me,
								"set Content-Transfer-Encoding", "hencoding",
								{encoding_parm}},
	{{ mime_set_charset}, "mime_set_charset", ti_mi,
								"set charset", "hcharset",
								{any_int_parm}},
	{{ mime_set_disposition}, "mime_set_disposition", ti_mi,
								"set Content-Disposition", "hdispo",
								{any_int_parm}},
	{{ mime_get_content}, "mime_get_content", tcc_mi, "get content", "gc",
								{any_int_parm}}, 
	{{ mime_get_available_buffer}, "mime_get_available_buffer", ti_m,
								"check space left in buffer", "buff",
								{no_parm}},
	{{ mime_clear_buffer}, "mime_clear_buffer", ti_m,
								"stop keeping content in buffer", "cbuff",
								{no_parm}},
	{{ mime_clear_boundary_all}, "mime_clear_boundary_all", ti_m,
								"clear all boundaries", "cboundary",
								{no_parm}},
	{{ mime_clear_encoding_all}, "mime_clear_encoding_all", ti_m,
								"clear all encodings", "cencode",
								{no_parm}},
	{{ mime_clear_content}, "mime_clear_content", ti_m,
								"discard entity content", "clear",
								{no_parm}},
	{{ line_of_content}, "mime_add_line_of_content", ti_mcc,
								"write a line of entity text",  "line",
								{any_string_parm}},
	{{ mime_add_file_of_content}, "mime_add_file_of_content", ti_mIi,
								"copy file into mime data", "file",
								{input_file_parm, any_int_parm}},
	{{ mime_attach_binary}, "mime_attach_binary", ti_mI,
								"attach a binary file", "attach",
								{input_file_parm}},
	{{ mime_get_depth}, "mime_get_depth", ti_m,
								"get depth of current pos", "depth",
								{no_parm}},
	{{ mime_get_position}, "mime_get_position", ti_m,
								"get current pos", "pos",
								{no_parm}},
	{{ mime_set_position}, "mime_set_position", ti_mi,
								"go to current pos", "goto",
								{any_int_parm}},
	{{ mime_seek}, "mime_seek", ti_mseek, "navigate the tree", "s",
								{seek_parm}},
	{{ test_set_header}, "mime_set_header", ti_mcc,
								"set header of current entity",  "h",
								{any_string_parm}},
	{{ mime_mark_for_discard}, "mime_mark_for_discard", ti_m, "mark for discard", "md",
								{no_parm}},
	{{ mime_commit_discard}, "mime_commit_discard", ti_m, "commit discard", "cd",
								{no_parm}},
	{{ mime_wrap}, "mime_wrap", ti_m, "wrap new entity around current", "wrap",
								{no_parm}},
	{{ mime_insert}, "mime_insert", ti_m, "insert new entity before current", "ins",
								{no_parm}},
	{{ mime_append}, "mime_append", ti_m, "append new entity after current", "app",
								{no_parm}},
	{{ mime_write}, "mime_write", ti_mfii, "write current entity to file", "write",
								{output_file_parm, unix_break_parm, dot_stuffing_parm}},
	{{ mime_extract_single_body}, "mime_extract_single_body", ti_mfeu,
								"write content only", "xtract",
								{output_file_parm, encoding_parm, unix_break_parm}},
	{{ mime_commit_encoding}, "mime_commit_encoding", te_me,
								"commit encoding", "encode",
								{encoding_parm}},

	{{ report_mime_all}, "", ti_mf_report, "report the whole mime tree", "+", {no_parm}},
	{{ report_mime_one}, "", ti_mf_report, "report the current entity", "=", {no_parm}},
	{{ report_mime_struct}, "", ti_mf_report, "report the mime structure", "*", {no_parm}},
	{{ leak_memory}, "", tv_v, "leak a small amount of memory", "leak", {no_parm}},
	{{ debug_fn_help}, "", tv_v, "display this help", "?", {no_parm}},
	{{ NULL}, "", tv, "quit", "q", {no_parm}}
};

unsigned char *get_command(unsigned char **s)
{
	unsigned char *t = *s, *p;
	
	while (t && *t && isspace(*t))
		++t;
	p = t;
	if (p && *p == 0)
		p = NULL;
		
	while (t && *t && !isspace(*t))
		++t;
	if (t && *t)
	{
		*t++ = 0;
		while (*t && isspace(*t))
			++t;
	}
	if (t && *t == 0)
		t = NULL;

	*s = t;
	return p;
}

int getintparm(unsigned char **s, int *result)
{
	char *t = NULL, *p;
	long l;
	
	if ((p = get_command(s)) != NULL &&
		(l = strtol(p, &t, 0)) <= INT_MAX &&
		l >= INT_MIN &&
		t && t > p && *t == 0)
	{
		*result = l;
		return 0;
	}

	if (*result != 0)
	{
		if (p)
			fprintf(stderr, "invalid integer parameter %s\n", p);
		else
			fprintf(stderr, "missing integer parameter\n");
	}
	
	return -1;
}

unsigned char *getstrparm(unsigned char **s)
{
	unsigned char *t = *s, *p;
	
	while (t && *t && isspace(*t))
		++t;
	p = t;
	while (t && *t)
		++t;
	if (p == t)
		p = NULL;
	*s = NULL;
	return p;
}

static const char encoding_syntax[] =
	"0 | base64 | qp | 7 | 8 | bin";
int getencodingparm(unsigned char **s, enum mime_encoding_enum *en)
{
	unsigned char *p = get_command(s);
	
	if (p)
	{
		if (stricmp(p, "0") == 0)
			*en = mime_encoding_not_set;
		else if (stricmp(p, "base64") == 0)
			*en = mime_base_64;
		else if (stricmp(p, "qp") == 0)
			*en = mime_quoted_printable;
		else if (stricmp(p,"7") == 0)
			*en = mime_7bit;
		else if (stricmp(p,"8") == 0)
			*en = mime_8bit;
		else if (stricmp(p,"bin") == 0)
			*en = mime_binary;
		else
		{
			fprintf(stderr, "invalid encoding parameter %s\n", p);
			return -1;
		}
		return 0;
	}
	fputs("missing encoding parmeter\n", stderr);
	return -1;
}

static const char seek_syntax[] =
	"top | downl | downr | right | left | up | first | last | next | prev";
int getseekparm(unsigned char **s, int *result)
{
	unsigned char *p = get_command(s);
	char buf[128], *b;
	
	if (p)
	{
		strcpy(buf, seek_syntax);
		for (b = strtok(buf, "| "), *result = 0;
			b != NULL;
			b = strtok(NULL, "| "), ++*result)
				if (stricmp(b, p) == 0)
					return 0;

		fprintf(stderr, "invalid seek parameter %s\n", p);
		return -1;
	}
	fputs("missing seek parmeter\n", stderr);
	return -1;
}

int getfileparm(unsigned char **s, int is_output, FILE **fp)
{
	char const *om[] = {"r+b", "w+b"};
	int rtc = -1;
	unsigned char *p = get_command(s);
	
	if (p && *p)
	{
		errno = 0;
		if (p[0] == '-' && p[1] == 0)
		{
			if (is_output)
				*fp = stdout;
			else
				*fp = stdin;
			rtc = 0;
		}
		else
		{
			*fp = fopen(p, om[is_output != 0]);
			if (*fp == NULL)
				fprintf(stderr, "cannot open \"%s\" (mode %s): %s\n",
					p, om[is_output != 0], strerror(errno));
			else
				rtc = 0;
		}
	}
	else
		fputs("missing file parameter\n", stderr);
	return rtc;
}

void debug_fn_help()
{
	unsigned i;
	FILE *outfp = stdout;
	char desc_hdr[] = "description",
		name_hdr[] = "function",
		key_hdr[] = "command";
	size_t desc_size = sizeof desc_hdr,
		name_size = sizeof desc_hdr,
		key_size = sizeof desc_hdr;

	for (i = 0; i < sizeof dbgtab/sizeof dbgtab[0]; ++i)
	{
		size_t ds = strlen(dbgtab[i].desc);
		size_t ns = strlen(dbgtab[i].name);
		size_t ks = strlen(dbgtab[i].key);
		
		if (ds > desc_size) desc_size = ds;
		if (ns > name_size) name_size = ns;
		if (ks > key_size) key_size = ks;
	}

	fprintf(outfp, "%-*s %-*s %s and param(s)\n",
		desc_size, desc_hdr,
		name_size, name_hdr,
		key_hdr);

	for (i = 0; i < sizeof dbgtab/sizeof dbgtab[0]; ++i)
	{
		unsigned j;
		fprintf(outfp, "%-*s %-*s %s",
			desc_size, dbgtab[i].desc,
			name_size, dbgtab[i].name,
			dbgtab[i].key);
		for (j = 0; j < sizeof dbgtab[0].parm / sizeof dbgtab[0].parm[0]; ++j)
			switch (dbgtab[i].parm[j])
			{
				case any_int_parm:
					fputs(" <int>", outfp);
					break;
					
				case seek_parm:
					fputs(" <seek>", outfp);
					break;
					
				case any_string_parm:
					fputs(" <string>", outfp);
					break;
					
				case no_parm:
					j = 10000;
					break;
					
				case output_file_parm:
					fputs(" <o file>", outfp);
					break;
					
				case input_file_parm:
					fputs(" <i file>", outfp);
					break;
					
				case encoding_parm:
					fputs(" <encoding>", outfp);
					break;
					
				case unix_break_parm:
					fputs(" <no-cr>", outfp);
					break;
				
				case dot_stuffing_parm:
					fputs(" <dot-st>", outfp);
					break;
				
				default:
					break;
			}
		fputs("\n", outfp);
	}
	fprintf( outfp,
		"---where:\n"
		"<string>   argument lasts until the end of the line\n"
		"<i file>   input file name, - for stdin\n"
		"<o file>   output file name, - for stdout\n"
		"<encoding> %s\n"
		"<no-cr>    unix line break: 0 for \"\\r\\n\" | 1 for \"\\n\"\n"
		"<dot-st>   dot-stuffing: 0 don't | 1 double leading dots on text lines\n"
		"<seek>     %s\n"
		"[repeat]	optional repeat count defaults to 1\n\n",
			encoding_syntax, seek_syntax);
}

char const *get_encoding_rtc(mime_encoding_rtc ertc)
{
	static char const *desc[] = 
	{
		"mime_ok_7bit", /* mime_identity + result domain */
		"mime_ok_8bit",
		"mime_ok_binary",
		
		/* mime_bad_result */
		"mime_bad_argument" /* = mime_bad_result */,
		"mime_bad_structure",
		"mime_cannot_do_boundary",
		"mime_system_error"
	};

	unsigned i = ertc;
	assert(i < sizeof desc/sizeof desc[0]);
	
	if (i < sizeof desc/sizeof desc[0])
		return desc[i];
	else
		return "INVALID mime_encoding_rtc";
}

char const *get_encoding_desc(enum mime_encoding_enum en)
{
	unsigned i = en;
	
	if (i < sizeof mime_encoding_desc/sizeof mime_encoding_desc[0])
		return mime_encoding_desc[i];
	else
		return "INVALID mime_encoding_enum";
}

void set_or_close(FILE *fp, file_param_array *fpa, int is_output)
{
	if (is_output)
		fflush(fp);
	if (fp != stdin && fp != stdout)
	{
		if (fpa->cur < sizeof fpa->fp / sizeof fpa->fp[0])
			fpa->fp[fpa->cur++] = fp;
		else
			fclose(fp);
	}
}

int run_debug_fn(MIME *m, int count, int *errs, FILE *outfp,
	file_param_array *fpa, char const *command)
{
	size_t i;
	char cmdbuf[1024];
	unsigned char *s, *cmd = NULL;
	int errcount = 0;
	int silent = be_silent;
	
	if (command != NULL)
	{
		if (strlen(command) + 4 >= sizeof cmdbuf)
			return -1;
		strcpy(cmdbuf, command);
		strcat(cmdbuf, "\n");
		s = cmdbuf;
		if (!silent)
			fprintf(outfp, "%d> %s", count, s);
	}
	else
	{
		fprintf(outfp, "%d> ", count);
		fflush(outfp);
		s = fgets(cmdbuf, sizeof cmdbuf, stdin);
		silent = 0;
	}
	if (s)
	{
		i = strlen(s);
		if (i > 0 && s[i-1] == '\n')
		{
			s[i-1] = 0;
			cmd = get_command(&s);
		}
	}
	if (cmd == NULL)
		return count;
	
	for (i = 0; i < sizeof dbgtab/sizeof dbgtab[0]; ++i)
	{
		if (stricmp(dbgtab[i].key, cmd) == 0)
		{
			union mime_fn_sig fn = dbgtab[i].fn;
			enum fn_type fn_t = dbgtab[i].fn_t;
			clock_t start, stop;
			int parm_ok = 1, ntimes = 0, nti = 0;

			int p1 = 1, p2 = 1, rtc = 0;
			FILE *fp;
			unsigned char *s1 = "";
			char const *rtcc = NULL;
			enum mime_encoding_enum encoding;
			mime_encoding_rtc ertc = mime_bad_result;
			
			if (fn.v == NULL)
				return -1;
			
			/*
			** get parameters
			*/
			switch(fn_t)
			{
				case ti_mi:
				case tcc_mi:
					parm_ok = getintparm(&s, &p1) == 0;
					break;

				case ti_mseek:
					parm_ok = getseekparm(&s, &p1) == 0;
					break;

				case ti_mcc:
					s1 = getstrparm(&s);
					parm_ok = s1 != NULL;
					break;
					
				case ti_mI:
					parm_ok = getfileparm(&s, 0, &fp) == 0;
					break;
						
				case ti_mIi:
					parm_ok = getfileparm(&s, 0, &fp) == 0 &&
						getintparm(&s, &p1) == 0;
					break;
					
				case ti_mfii:
					parm_ok = getfileparm(&s, 1, &fp) == 0 &&
						getintparm(&s, &p1) == 0 &&
						getintparm(&s, &p2) == 0;
					break;
						
				case ti_mfeu:
					parm_ok = getfileparm(&s, 1, &fp) == 0 &&
						getencodingparm(&s, &encoding) == 0 &&
						getintparm(&s, &p1) == 0;
					break;
						
				case te_me:
				case ti_me:
					parm_ok = getencodingparm(&s, &encoding) == 0;
					break;
					
				default:
					break;
			}
			
			/*
			** get repeat count
			*/
			if (parm_ok && getintparm(&s, &ntimes) != 0)
				ntimes = 1;
			
			/*
			** run as many times or until error
			*/
			if (parm_ok)
			{
				clock();
				start = clock();
				for (nti = 0; nti < ntimes && errcount == 0; ++nti)
					switch(fn_t)
					{
						case tv_v: /* debug fn help - no count */
							(*fn.v_v)();
							break;
		
						case ti_m:
							rtc = (*fn.i_m)(m);
							if (rtc == -1)
								++errcount;
							break;
							
						case ti_mi:
							rtc = (*fn.i_mi)(m, p1);
							if (rtc < 0)
								++errcount;
							break;
		
						case ti_mseek:
							rtc = (*fn.i_mi)(m, p1);
							if (rtc < 0)
								++errcount;
							break;
		
						case tcc_m:
							rtcc = (*fn.cc_m)(m);
							break;
		
						case tcc_mi:
							rtcc = (*fn.cc_mi)(m, p1);
							if (rtcc == NULL)
								++errcount;
							break;
						
						case ti_mcc:
							rtc = (*fn.i_mcc)(m, s1);
							if (rtc == -1)
								++errcount;
							break;
							
						case ti_mf_report: /* report fn */
							(*fn.i_mf_report)(m, outfp);
							break;
							
						case ti_mI:
							rtc = (*fn.i_mI)(m, fp);
							if (rtc == -1)
								++errcount;
							break;
								
						case ti_mIi:
							rtc = (*fn.i_mIi)(m, fp, p1);
							if (rtc != 0)
								++errcount;
							break;
							
						case ti_mfii:
							rtc = (*fn.i_mfii)(m, fp, p1, p2);
							if (rtc != 0)
								++errcount;
							break;
								
						case ti_mfeu:
							rtc = (*fn.i_mfeu)(m, fp, encoding, p1);
							if (rtc != 0)
								++errcount;
							break;
								
						case te_me:
							ertc = (*fn.e_me)(m, encoding);
							if (ertc >= mime_bad_result)
								++errcount;
							break;
							
						case ti_me:
							rtc = (*fn.e_me)(m, encoding);
							if (rtc != 0)
								++errcount;
							break;
							
						default:
							assert(0);
							break;
					}
				stop = clock();
			
				/*
				** display last result, add to count, close or set files
				*/
				if (errcount && ntimes > 1)
					fprintf(outfp, "ERROR after %d/%d times\a\n", nti, ntimes);
				
				switch(fn_t)
				{
					case ti_m:
					case ti_mi:
					case ti_mseek:
					case ti_mcc:
					case ti_me:
						if (!silent)
							fprintf(outfp, "%d", rtc);
						++count;
						break;
						
					case tcc_m:
					case tcc_mi:
						if (!silent)
							fprintf(outfp, "\"%s\"", rtcc ? rtcc : "NULL");
						++count;
						break;
					
					case ti_mI:
					case ti_mIi:
					case ti_mfii:
					case ti_mfeu:
						if (!silent)
							fprintf(outfp, "%d", rtc);
						set_or_close(fp, fpa, 0);
						++count;
						break;
							
					case te_me:
						if (!silent || ertc >= mime_bad_result)
							fprintf(outfp, "%s", get_encoding_rtc(ertc));
						++count;
						break;
						
					default:
						parm_ok = 0;
						break;
				}
				if (parm_ok && !silent)
					fprintf(outfp, " from %s [%d/%d time(s) in %ldms]\n",
						dbgtab[i].name, nti, ntimes, (long)(stop - start)/1000);
			}
							
			break;
		}
	}
	
	if (i >= sizeof dbgtab/sizeof dbgtab[0])
		fprintf(outfp, "%s not found, try '?'\n", Psz(s));
	else if (s && *s && !silent)
		fprintf(outfp, "extra parameters discarded: \"%s\"\n", s);

	*errs = errcount;
	return count;
}

int cdispl(char *dest, char *edest, char *src, char *esrc)
{
	char *buf, *pb = src, *pc;
	int dsize, ssize;

	assert(dest && edest && src && esrc);
	assert(dest + 10 < edest);
	assert(src <= esrc);

	dsize = edest - dest;
	memset(dest, 0, dsize);

	buf = pc = (char*) malloc(4*(esrc - src) + 1); /* brute force */
	if (buf == NULL)
		return 0;

	while (pb < esrc)
	{
		int ch = *(unsigned char*)pb++;
		if (isalnum(ch) || ispunct(ch) || ch == ' ')
		{
			*pc++ = ch;
			if (ch == '\\')
				*pc++ = ch;
		}
		else
		{
			*pc++ = '\\';
			switch (ch)
			{
				case '\r': *pc++ = 'r'; break;
				case '\n': *pc++ = 'n'; break;
				case '\t': *pc++ = 't'; break;
				case '\a': *pc++ = 'a'; break;
				case '\f': *pc++ = 'f'; break;
				default:
					pc += sprintf(pc, "x%02X", ch);
					break;
			}
		}
	}

	ssize = pc - buf;
	if (ssize < dsize)
		memcpy(dest, buf, ssize);
	else
	{
		dsize /= 2;
		memcpy(dest, buf, dsize);
		memcpy(edest - dsize - 1, pc - dsize, dsize);
		memcpy(dest + dsize - 1, "...", 3);
	}

	free(buf);
	return ssize; // if larger than dest, dots...
}

void cdispl_lines(FILE* fp, char *src, char *esrc)
{
	static int const per_buf = 20, line_length = 75, max_lines = 5;
	char buf[80]; // per_buf * 4;
	int l = 0, finish = 0;
	int ssize = line_length + 1;
	unsigned off = 0;
	while (src < esrc)
	{
		char *e = src + per_buf;
		if ( e > esrc)
			e = esrc;
		if (ssize > line_length)
		{
			if (++l >= max_lines && !finish)
			{
				char *resu = esrc - max_lines*line_length/4;
				finish = 1;
				if (resu - line_length > e)
				{
					off += resu - src;
					src = resu;
					e = src + per_buf;
					if (e > esrc)
						e = esrc;
					fputs("...", fp);
				}
			}
			ssize = fprintf(fp, "\n%8u ", off);
		}
		ssize += cdispl(buf, buf + sizeof buf, src, e);
		off += e - src;
		src = e;
		fputs(buf, fp);
	}
	fputc('\n', fp);	
}

void report_file(FILE *outfp, FILE*fp, char const * const desc)
{
	if (be_silent)
		fputs(desc, outfp);
	else
	{
		fprintf(outfp, "%s: ", desc);
		if (fp)
		{
			fprintf(outfp, "handle %d%s%s", fileno(fp),
				feof(fp) ? " feof" : "", ferror(fp) ? " ferror" : "");
		}
		else
			fputs("NULL", outfp);
	}
}

void report_mime_struct(MIME *m, FILE *outfp)
{
	static const char* desc[3] = { "buffer", "buf_first", "buf_end"};
	char *ptr[4];
	char content[41];
	char *nextptr;
	int i;

	if (mime_get_last_error(m))
		fprintf(outfp, "ERROR %d: %s\n",
			mime_get_last_error(m), mime_get_last_error_string(m));
	report_file(outfp, m->fp, "fp");
	putc('\n', outfp);
	report_file(outfp, m->fpw, "fpw");
	putc('\n', outfp);

	fprintf(outfp, "next_internal_id=%d\n", m->next_internal_id);
	fprintf(outfp, "ment_max=%u, ment_cur=%u, ment_count=%u, ment_committed=%u\n",
		m->ment_max, m->ment_cur, m->ment_count, m->ment_committed);
	fprintf(outfp, "default_encoding=%u, default_charset=%u\n"
		"max_data_size=%ld, buf_max=%u\n",
		m->default_encoding, m->default_charset, m->max_data_size, m->buf_max);

	cdispl(content, content + sizeof content, m->crlf, m->crlf + sizeof m->crlf);
	fprintf(outfp, "%u+%u line break=%s, crlf=%s",
		m->header_u_count, m->header_b_count,
		m->unix_line_break ? "unix" : "standard", content);
	if (m->dot_stuffing) fputs(" dot_stuffing", outfp);
	if (m->seek_stuck) fputs(" seek_stuck", outfp);
	fputs("\n\n", outfp);

	ptr[0] = m->buffer;
	ptr[1] = m->buf_first;
	ptr[2] = m->buf_end;
	ptr[3] = m->buffer + m->buf_max;
	for (i = 0; i < 3; ++i)
	{
		nextptr = ptr[i+1];
		cdispl(content, content + sizeof content, ptr[i], nextptr);
		printf("%8p (%4d)  %-16s %s\n",
			ptr[i], nextptr - ptr[i], desc[i], content);
	}
	fputs("\n", outfp);

	fprintf(outfp, "boundary_ndx=%u\n", m->boundary_ndx);
	cdispl(content, content + sizeof content,
		m->boundary_rand, m->boundary_rand + sizeof m->boundary_rand);
	fprintf(outfp, "boundary_rand=%s\n", content);
	cdispl(content, content + sizeof content,
		m->boundary_unique, m->boundary_unique + sizeof m->boundary_unique);
	fprintf(outfp, "boundary_unique=%s\n", content);
	fprintf(outfp, "no_boundary=0x%08lx  0x%08lx\n",
		m->no_boundary[0], m->no_boundary[1]);
	fputs("\n", outfp);
}

typedef struct debug_mime_type
{
	unsigned short int type_enum;
	char *descr;
} debug_mime_type;

static const debug_mime_type mime_type_descr[] =
{
	{ 	mime_text, "text"},
  /* mime_image, mime_audio, mime_video,*/
   { mime_application, "application"},
	{ mime_multipart, "multipart"},
	{ mime_message, "message"},
	{ 999, NULL}
};

static const debug_mime_type mime_subtype_descr[] =
{
	{ mime_text_plain, "plain"},
	{ mime_text_other, "other text"},
	{ mime_application_octetstream, "octet-stream"},
	{ mime_multipart_mixed, "mixed"},
	{ mime_multipart_form_data, "form-data"},
	{ mime_message_rfc822, "rfc822"},
	{ mime_message_rfc822_headers, "rfc822 (headers only)"},
	{ mime_message_machine, "\"machine\""},
	{ mime_message_other, "other message"},
	{ 999, NULL}
};

static void get_type_descr(unsigned type_enum, debug_mime_type const* tbl, char *desc)
{
	int i;
	
	for (i = 0; tbl[i].descr != NULL; ++i)
		if (tbl[i].type_enum == type_enum)
		{
			strcpy(desc, tbl[i].descr);
			return;
		}
	sprintf(desc, "unknown-%d", type_enum);
}

int report_mime(MIME *m, FILE *fp, size_t i, int indent, int all)
{
	for (; i < m->ment_count; ++i)
	{
		MIME_ENTITY* const ment = m->ment + i;
		char type_descr[40], subtype_descr[40];

		get_type_descr(ment->mime_type, mime_type_descr, type_descr);
		get_type_descr(ment->mime_subtype, mime_subtype_descr, subtype_descr);
		
		/* first (possibly only) line */
		fprintf(fp, "%*s%02d: _%d_ %d", indent, "", i,
			ment->internal_id, get_level(m, i));
		if (ment->error_base64)
			fputs(" ERR-base64", fp);
		if (ment->error_quoted_printable)
			fputs(" ERR-quoted-printable", fp);
		if (ment->error_system)
			fputs(" ERR-system", fp);
		if (ment->is_leaf_part && !ment->found_term_boundary)
			fputs(" no-term-boundary", fp);
		if (ment->description)
			fprintf(fp, " \"%s\"", ment->description);
		fprintf(fp, " %s/%s", type_descr, subtype_descr);
		if (ment->content_type)
			fprintf(fp, " (\"%s\")", ment->content_type);
		if (ment->encoding)
			fprintf(fp, " enc=%s", get_encoding_desc(ment->encoding));
		if (ment->actual_encoding)
			fprintf(fp, " act=%s", get_encoding_desc(ment->actual_encoding));
		if (ment->charset == 1)
			fputs(" us-ascii", fp);
		else if (ment->charset > 1)
			fprintf(fp, " iso-8859-%d", ment->charset - 1);
		else if (all == 0)
			fputs(" no-charset", fp);
		if (ment->fpw || all == 0)
			report_file(fp, ment->fpw, " fpw");
		if (m->ment_cur == i)
			fputs(" <=====CUR", fp);
		fputs("\n", fp);
			
		/* flags line (?) */
		if (all == 0 ||
			ment->dispo != mime_dispo_not_set ||
			ment->is_alt_file ||
			ment->is_for_discard ||
			ment->is_last_part ||
			ment->is_leaf_part == 0 ||
			ment->is_discrete == 0 ||
			ment->error_base64 ||
			ment->error_quoted_printable ||
			ment->error_system ||
			ment->error_structure ||
			ment->warning_structure)
		{
			char const *none = all ? "" : " -";
			fprintf(fp, "%*s   %s%s%s%s%s%s%s%s%s%s%s%s%s\n", indent, "",
				ment->is_last_part ? " last" : none,
				ment->is_leaf_part ? none : " node",
				ment->is_discrete ? none : " container",
				ment->dispo == mime_dispo_is_form_data_field ? " field" : none,
				ment->dispo == mime_dispo_is_inline ? " inline" : none,
				ment->dispo == mime_dispo_is_attachment ? " attachment" : none,
				ment->is_alt_file ? " alt" : "",
				ment->is_for_discard ? " discard" : none,
				ment->error_base64 ? " error-base64" : none,
				ment->error_quoted_printable ? " error-quoted_printable" : none,
				ment->error_system ? " error-system" : none,
				ment->error_structure ? " error-structure" : none,
				ment->warning_structure ? " warning-structure" : none
				);
		}


		/* data line (?) */
		if (ment->length || ment->hlength || ment->actual_max_line_length ||
			ment->boundary || ment->name || ment->filename || all == 0)
		{
			fprintf(fp, "%*s   ", indent, "");
			if (ment->length || all == 0)
				fprintf(fp, " data offset=%ld length=%ld", 
					ment->offset, ment->length);
			if (ment->hlength || all == 0)
				fprintf(fp, " header offset=%ld length=%ld", 
					ment->hoffset, ment->hlength);
			if (ment->actual_max_line_length || all == 0)
				fprintf(fp, " max-l=%ld", ment->actual_max_line_length);
			if (ment->filename || all == 0)
				fprintf(fp, " file=\"%s\"", Psz(ment->filename));
			if (ment->name || all == 0)
				fprintf(fp, " name=\"%s\"", Psz(ment->name));
			if (ment->boundary || all == 0)
				fprintf(fp, " %s", be_silent? "boundary" : Psz(ment->boundary));
			fputs("\n", fp);
		}

		if (all == 0) // reporting only one
			break;
		if (!ment->is_leaf_part)
			i = report_mime(m, fp, i + 1, indent + 4, all);
		if (ment->is_last_part)
			return i;
	}
	
	if (all == 0)
	{
		MIME_ENTITY* const ment = m->ment + i;

		if (ment->hlength > 0)
		{
		}
		
		if (ment->length > 0 && ment->is_discrete)
		{
			content_ptr part;
			
			if (mime_mmap_content(m, &part) == 0)
			{
				cdispl_lines(fp, part.start, part.start + part.length);
				mime_unmap(m);
			}
			else
				fprintf(fp, "Cannot mmap MIME part (%s): %s\n",
					mime_get_last_error_string(m),
					strerror(errno));
		}
	}
	return i;
}

static int mime_testread(MIME* m, FILE* fp)
/*
** utility function "rdf" for reading a whole mime file
*/
{
	int rtc;
	long diff;
	FILE *fpold = m->fp;
	m->fp = fp;
	/*
	** reset buffer before read; this discards any
	** content previously kept there via mime_get_content
	*/
	m->buf_first = m->buf_end = m->buffer;
	m->ment_committed = -1;
	m->unix_line_break = 0;
	m->header_u_count = m->header_b_count = 0;
	while ((rtc = read_mime_header(m, 1)) > 0)
		continue;
	
	m->fp = fpold;
	diff = (long)(m->buf_end) - (long)(m->buf_first);
	if (diff > 0) /* this amount to be read again */
		fseek(fp, -diff, SEEK_CUR);

	if (rtc == 0)
		rtc = mime_read(m, fp);
	return rtc;
}

void usage(char const *s)
{
	fprintf(stderr, "usage: %s [<default-encoding> [<default-charset>]]\n"
		"where default encoding is %s\n"
		"and default charset is 0 for us-ascii, n for iso-8859-n [n=1, 15]\n",
			s, encoding_syntax);
}

int main(int argc, char *argv[])
/*
** this function is compiled conditionally to obtain a test program.
** this allows debugging each exposed function, and also provides a
** mean to interactively check the effect of any given sequence of
** calls. with some perverse bias, one might call it a mime editor
** with a bare-bones command-line style user interface.
**
** see usage function far parameters that belong to the MIME object
** used for a session. after the object has been successfully created
** one can submit various commands after a numbered prompt. the command
** line is interpreted in run_debug_fn above. entering a question mark
** gives the list of available functions, as defined in dbgtab above.
*/
{
	unsigned err = 0;
	int nextarg = 1;
	enum mime_encoding_enum default_encoding = mime_encoding_not_set;
	enum mime_charset_enum default_charset = mime_charset_not_set;
	
	for (nextarg = 1; nextarg < argc; ++nextarg)
	{
		char const *const arg = argv[nextarg];
		if (arg[0] != '-')
			break;
		else
		{
			if (strcmp(arg, "--silent") == 0 || strcmp(arg, "-s") == 0)
				be_silent = 1;
		}
	}
	
	if (argc > nextarg)
	{
		unsigned char *s = argv[nextarg];
		if (getencodingparm(&s, &default_encoding) != 0)
		{
			++err;
		}
		++nextarg;
		
		if (argc > nextarg)
		{
			char *t;
			unsigned long ul = strtoul(argv[nextarg], &t, 10);
			if (ul <= 15 && t && *t == 0)
				default_charset = (enum mime_charset_enum)(ul + 1);
			else
			{
				fprintf(stderr, "invalid charset %s\n", argv[2]);
				++err;
			}
			++nextarg;
		}
	}
	
	if (err)
	{
		usage(argv[0]);
		err = 1;
	}
	else
	{
		MIME *m = mime_new(default_encoding, default_charset);
		
		if (m)
		{
			int rtc = 1, i;
			file_param_array fpa;
			int errcount = 0;
			
			memset(&fpa, 0, sizeof fpa);
			
			for (i = nextarg; i < argc; ++i)
				rtc = run_debug_fn(m, rtc, &errcount, stdout, &fpa, argv[i]);
			for (; rtc > 0;)
				rtc = run_debug_fn(m, rtc, &errcount, stdout, &fpa, NULL);
			if (errcount > 0)
			{
				fprintf(stdout, "%d error(s) total\n", errcount);
				err = 2;
			}
			for (i = 0; (size_t)i < sizeof fpa.fp / sizeof fpa.fp[0]; ++i)
				if (fpa.fp[i] != NULL)
					fclose(fpa.fp[i]);
			
			mime_clear(m);
		}
		else
		{
			err = 2;
		}
	}
	
	return err; /* 0=Ok, 1=bad arg, 2=module errs, 3=abort */
}

#endif /* TEST_MAIN */
